# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2024, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from ._unit import Unit
from ._facility import Facility

__all__ = ('Module', 'FacilityModule')

class Module(Unit):
    _N_ins = 0
    _N_outs = 0
    _ins_size_is_fixed = False
    _outs_size_is_fixed = False
    
    def _assert_compatible_property_package(self): pass
    
    def _setup(self):
        self.auxiliary_system._setup(load_configuration=False)
        super()._setup()
    
    def _run(self):
        self.auxiliary_system.converge()
    
    def _design(self):
        for i in self.auxiliary_system.units: i._design()

    def _cost(self):
        for i in self.auxiliary_system.units: i._cost()
        
class FacilityModule(Module, Facility): network_priority = None


# -*- coding: utf-8 -*-
# Bioindustrial-Park: BioSTEAM's Premier Biorefinery Models and Results
# Copyright (C) 2021-2024, Yalin Li <mailto.yalin.li@gmail.com>
# Copyright (C) 2023, Yoel Rene Cortes-Pena <yoelcortes@gmail.com>
#
# This module is under the UIUC open-source license. See
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.

"""
This module contains data and utility functions for the high-rate wastewater treatment system.
"""

import biosteam as bst
import numpy as np
from chemicals.elements import molecular_weight
from warnings import warn
from thermosteam import Chemical, Chemicals, Stream, settings, Thermo
from thermosteam.reaction import (
    Reaction as Rxn,
    ParallelReaction as PRxn
)
mix_tank_purchase_cost_algorithms = bst.design_tools.mix_tank_purchase_cost_algorithms
TankPurchaseCostAlgorithm = bst.design_tools.TankPurchaseCostAlgorithm
ExponentialFunctor = bst.utils.ExponentialFunctor
remove_undefined_chemicals = bst.utils.remove_undefined_chemicals
default_chemical_dict = bst.utils.default_chemical_dict

__all__ = (
    # Combustion
    'get_combustion_energy',
    # Construction
    'IC_purchase_cost_algorithms',
    'select_pipe',
    'cost_pump',
    # Chemicals
    'default_insolubles',
    'get_insoluble_IDs',
    'get_soluble_IDs',
    'append_wwt_chemicals',
    'create_missing_wwt_chemicals',
    # Digestion
    'get_BD_dct',
    'get_digestable_chemicals',
    'compute_stream_COD',
    'get_digestion_rxns',
    # Miscellaneous
    'format_str',
    'remove_undefined_chemicals',
    'get_split_dct',
    'kph_to_tpd',
    # TEA/LCA
    'prices',
    'GWP_CFs',
    )


# %%

# =============================================================================
# Combustion
# =============================================================================

combustion_IDs = ['O2', 'H2O', 'CO2', 'N2', 'SO2', 'P4O10', 'Ash']
def create_combustion_chemicals():
    chems = Chemicals(combustion_IDs[:-1])
    for chem in chems:
        if chem.ID != 'P4O10': chem.at_state('g')
        else: chem.at_state('s')
    chems.append([Chemical('Ash', phase='s', phase_ref='s', search_db=False, MW=1.)])
    return chems


def get_combustion_energy(stream, combustion_eff=0.8):
    """
    Estimate the amount of energy generated from combustion of incoming streams.
    """
    chems = stream.chemicals
    to_add = []
    
    for ID in combustion_IDs:
        if not hasattr(chems, ID): to_add.append(ID)
    if to_add:
        combustion_chems = create_combustion_chemicals()
        for ID in to_add:
            to_add.append(combustion_chems.pop(ID))
        new_chems = Chemicals(to_add)
        new_chems.compile()
        settings.set_thermo(new_chems)
        new_stream = Stream()
        new_stream.imass[chems.IDs] = stream.imass[chems.IDs]
    else:
        new_chems = chems
        new_stream = stream
    rxns = new_stream.chemicals.get_combustion_reactions()
    reacted = new_stream.copy()
    rxns.force_reaction(reacted.mol)
    reacted.imol['O2'] = 0
    H_net = new_stream.H + new_stream.HHV - reacted.H
    settings.set_thermo(chems)
    return H_net


# %%

# =============================================================================
# Construction
# =============================================================================

##### Internal Circulation Reactor #####
IC_purchase_cost_algorithms = mix_tank_purchase_cost_algorithms.copy()
conventional = IC_purchase_cost_algorithms['Conventional']
# The cost correlation might not hold for the ranges beyond
ic = TankPurchaseCostAlgorithm(
    ExponentialFunctor(A=conventional.f_Cp.A,
                       n=conventional.f_Cp.n),
    V_min=np.pi/4*1.5**2*16, # 1.5 and 16 are the lower bounds of the width and height ranges in Kontos
    V_max=np.pi/4*12**2*25, # 12 and 25 are the lower bounds of the width and height ranges in Kontos
    V_units='m^3',
    CE=conventional.CE,
    material='Stainless steel')

IC_purchase_cost_algorithms['IC'] = ic

##### Pipe Selection #####
# Based on ANSI (American National Standards Institute) pipe chart
# the original code has a bug (no data for 22) and has been fixed here
boundaries = np.concatenate([
    np.arange(1/8, 0.5, 1/8),
    np.arange(0.5, 1.5, 1/4),
    np.arange(1.5, 5,   0.5),
    np.arange(5,   12,  1),
    np.arange(12,  36,  2),
    np.arange(36,  54,  6)
    ])

size = boundaries.shape[0]

pipe_dct = {
    1/8 : (0.405,  0.049), # OD (outer diameter), t (wall thickness)
    1/4 : (0.540,  0.065),
    3/8 : (0.675,  0.065),
    1/2 : (0.840,  0.083),
    3/4 : (1.050,  0.083),
    1   : (1.315,  0.109),
    1.25: (1.660,  0.109),
    1.5 : (1.900,  0.109),
    2   : (2.375,  0.109),
    2.5 : (2.875,  0.120),
    3   : (3.500,  0.120),
    3.5 : (4.000,  0.120),
    4   : (4.500,  0.120),
    4.5 : (5.000,  0.120),
    5   : (5.563,  0.134),
    6   : (6.625,  0.134),
    7   : (7.625,  0.134),
    8   : (8.625,  0.148),
    9   : (9.625,  0.148),
    10  : (10.750, 0.165),
    11  : (11.750, 0.165),
    12  : (12.750, 0.180),
    14  : (14.000, 0.188),
    16  : (16.000, 0.199),
    18  : (18.000, 0.188),
    20  : (20.000, 0.218),
    22  : (22.000, 0.250),
    24  : (24.000, 0.250),
    26  : (26.000, 0.250),
    28  : (28.000, 0.250),
    30  : (30.000, 0.312),
    32  : (32.000, 0.312),
    34  : (34.000, 0.312),
    36  : (36.000, 0.312),
    42  : (42.000, 0.312),
    48  : (48.000, 0.312)
    }


def select_pipe(Q, v):
    """Select pipe based on Q (flow in ft3/s) and velocity (ft/s)"""
    A = Q / v # cross-section area
    d = (4*A/np.pi) ** 0.5 # minimum inner diameter, [ft]
    d *= 12 # minimum inner diameter, [in]
    d_index = np.searchsorted(boundaries, d, side='left') # a[i-1] < v <= a[i]
    d_index = d_index-1 if d_index==size else d_index # if beyond the largest size
    OD, t = pipe_dct[boundaries[d_index]]
    ID = OD - 2*t # inner diameter, [in]
    return OD, t, ID


##### Pumping #####
def cost_pump(unit):
    """
    Calculate the cost of the pump and pump building for a unit
    based on its `Q_mgd` (hydraulic flow in million gallons per day),
    `recir_ratio` (recirculation ratio) attributes.
    """

    Q_mgd, recir_ratio = unit.Q_mgd, unit.recir_ratio

    # Installed pump cost, this is a fitted curve
    pumps = 2.065e5 + 7.721*1e4*Q_mgd

    # Design capacity of intermediate pumps, gpm,
    # 2 is the excess capacity factor to handle peak flows
    GPMI = 2 * Q_mgd * 1e6 / 24 / 60

    # Design capacity of recirculation pumps, gpm
    GPMR = recir_ratio * Q_mgd * 1e6 / 24 / 60

    building = 0.
    for GPM in (GPMI, GPMR):
        if GPM == 0:
            N = 0
        else:
            N = 1 # number of buildings
            GPMi = GPM
            while GPMi > 80000:
                N += 1
                GPMi = GPM / N

        PBA = N * (0.0284*GPM+640) # pump building area, [ft]
        building += 90 * PBA

    return pumps, building


# %%

# =============================================================================
# Chemicals
# =============================================================================

#!!! This should be individually added when adding the WWT process
default_insolubles = {
    # Sugarcane
    'CaO', 'Flocculant', 'Solids', 'Yeast',
    # Corn
    'Fiber', 'InsolubleProtein',
    # Corn stover
    'Ash', 'CaSO4', 'Cellulose', 'DenaturedEnzyme', 'Enzyme', 'Hemicellulose',
    'Lignin', 'Lime', 'P4O10', 'Protein', 'Tar', 'T_reesei', 'WWTsludge', 'Z_mobilis',
    # Lactic acid
     'Arabinan', 'Acetate', 'BaghouseBag', 'CalciumDihydroxide', 'CoolingTowerChems',
     'FermMicrobe', 'Galactan', 'Glucan','Mannan', 'Polymer', 'Xylan',
    # Oilcane
    'Biomass', 'Cellmass', 'DryYeast',
    # Others
    'CellMass',
    }

def get_insoluble_IDs(chemicals, insolubles):
    chem_IDs = set([i.ID for i in chemicals])
    new_insolubles = set(insolubles).intersection(chem_IDs)
    return tuple(new_insolubles)

def get_soluble_IDs(chemicals, insolubles):
    return tuple([i.ID for i in chemicals if not i.ID in insolubles])


_cal2joule = 4.184 # auom('cal').conversion_factor('J')

def create_missing_wwt_chemicals(chemicals):
    new_chemicals = Chemicals([])
    def add_chemical(ID, **data):
        missing = not hasattr(chemicals, ID)
        if missing:
            chemical = Chemical(ID, **data)
            new_chemicals.append(chemical)
        return missing

    add_chemical('NH3', phase='g', Hf=-10963*_cal2joule)
    add_chemical('H2S', phase='g', Hf=-4927*_cal2joule)
    add_chemical('SO2', phase='g')
    add_chemical('NH4OH', search_ID='AmmoniumHydroxide', phase='l', Hf=-336719)
    add_chemical('H2SO4', phase='l')
    add_chemical('HCl', phase='l')
    add_chemical('HNO3', phase='l', Hf=-41406*_cal2joule)
    add_chemical('NaOH', phase='l'),
    add_chemical('NaNO3', phase='l', Hf=-118756*_cal2joule)
    add_chemical('Na2SO4', phase='l', Hf=-1356380)
    if add_chemical('CaSO4', phase='s', Hf=-342531*_cal2joule):
        new_chemicals.CaSO4.Cn.move_up_model_priority('LASTOVKA_S', 0)
    add_chemical('NaOCl', phase='l', Hf=-347*1e3) # https://en.wikipedia.org/wiki/Sodium_hypochlorite
    add_chemical('CitricAcid', phase='l', Hf=-1543.8*1e3) # https://en.wikipedia.org/wiki/Citric_acid
    add_chemical('Bisulfite', phase='l')
    add_chemical('WWTsludge', search_db=False, phase='s',
                formula='CH1.64O0.39N0.23S0.0035', Hf=-23200.01*_cal2joule)
    if add_chemical('Polymer', search_db=False, phase='s', MW=1, Hf=0, HHV=0, LHV=0):
        new_chemicals.Polymer.Cn.add_model(evaluate=0, name='Constant')
    for i in new_chemicals: i.default()
    return new_chemicals


def append_wwt_chemicals(chemicals, set_thermo=True):
    chemicals = chemicals.chemicals if isinstance(chemicals, Thermo) else chemicals
    required_chemicals = (
        'NH3', 'H2S', 'SO2', 'NH4OH', 'H2SO4', 'HCl', 'HNO3', 'NaOH', 'NaNO3',
        'Na2SO4', 'CaSO4', 'NaOCl', 'CitricAcid', 'Bisulfite', 'WWTsludge', 'Polymer'
    )
    if all([(i in chemicals) for i in required_chemicals]): return chemicals
    chems = Chemicals([*chemicals, *create_missing_wwt_chemicals(chemicals)])
    chems.compile()

    # Add aliases and groups
    get = getattr
    for grp, comp in chemicals._group_mol_compositions.items():
        group_IDs = [chem.ID for chem in get(chemicals, grp)]
        chems.define_group(grp, group_IDs, comp)

    if set_thermo: settings.set_thermo(chems)
    return chems



# %%

# =============================================================================
# Digestion
# =============================================================================

def get_CHONSP(chemical):
    organic = True
    atoms = chemical.atoms

    CHONSP = []
    for atom in ('C', 'H', 'O', 'N', 'S', 'P'):
        CHONSP.append(atoms.get(atom) or 0.,)

    if CHONSP[0] <= 0 or CHONSP[1] <= 0: # does not have C or H
        if not (len(atoms) == 1 and CHONSP[1] == 2): # count H2 as organic
            organic = False

    if sum(v for v in atoms.values()) != sum(CHONSP): # contains other elements
        organic = False

    return CHONSP if organic else [0.]*6


def get_COD_stoichiometry(chemical):
    r"""
    Get the molar stoichiometry for the theoretical
    chemical oxygen demand (COD) of a given chemical as in:

    .. math::
        C_nH_aO_bN_cS_dP_e + \frac{2n+0.5a-b-1.5c+3d+2.5e}{2}O_2
        -> nCO_2 + \frac{a-3c-2d}{2}H_2O + cNH_3 + dH_2SO_4 + \frac{e}{4}P_4O_{10}
    """
    Xs = nC, nH, nO, nN, nS, nP = get_CHONSP(chemical)

    excluded = ('O2', 'CO2', 'H2O', 'NH3', 'H2SO4', 'P4O10',
                *default_insolubles)
    if chemical.ID in excluded or chemical.locked_state=='g':
        return dict.fromkeys(excluded, 0)

    dct = {
        chemical.ID: -1. if sum([abs(i) for i in Xs])!=0 else 0.,
        'O2': -(nC+0.25*nH-0.5*nO-0.75*nN+1.5*nS+1.25*nP),
        'CO2': nC,
        'H2O': 0.5*nH-1.5*nN-nS, # assume one water reacts with SO3 to H2SO4
        'NH3': nN,
        'H2SO4': nS,
        'P4O10': 0.25*nP
        }

    return dct


def get_digestable_chemicals(chemicals):
    chems = [chemicals[i.ID] for i in chemicals
             if get_COD_stoichiometry(i)['O2']!=0]
    return chems


def get_BMP_stoichiometry(chemical):
    r"""
    Compute the theoretical biochemical methane potential (BMP) in
    mol :math:`CH_4`/mol chemical of a given chemical using:

    .. math::
        C_vH_wO_xN_yS_z + \frac{4v-w-2x+3y+2z}{2}H2O ->
        \frac{4v+w-2x-3y-2z}{8}CH4 + \frac{(4v-w+2x+3y+2z)}{8}CO2 + yNH_3 + zH_2S
    """
    Xs = nC, nH, nO, nN, nS, nP = get_CHONSP(chemical)

    excluded = ('H2O', 'CH4', 'CO2', 'NH3', 'H2S',
                *default_insolubles)
    if chemical.ID in excluded or chemical.locked_state=='g':
        return dict.fromkeys(excluded, 0)

    dct = {
        chemical.ID: -1. if sum([abs(i) for i in Xs])!=0 else 0.,
        'H2O': -(nC-0.25*nH-0.5*nO+0.75*nN+0.5*nS),
        'CH4': 0.5*nC+0.125*nH-0.25*nO-0.375*nN-0.25*nS,
        'CO2': 0.5*nC-0.125*nH+0.25*nO+0.375*nN+0.25*nS,
        'NH3': nN,
        'H2S': nS,
        }

    return dct


# Note that these biodegradabilities will then be multiplied by the yield
# of biogas/cell mass (0.86 is the default)
def get_BD_dct(chemicals, default_BD=1, **kwargs):
    BD_dct = dict.fromkeys([i.ID for i in get_digestable_chemicals(chemicals)],
                           default_BD)

    # Based on Kontos
    BD_dct['AceticAcid'] = 1 # 0.87/0.86>1
    BD_dct['Arabinose'] = 0.2/0.86
    BD_dct['Glucose'] = 1 # 0.87/0.86>1
    BD_dct['GlucoseOligomer'] = BD_dct['Glucan'] = 0.81/0.86
    BD_dct['HMF'] = 0.85/0.86
    BD_dct['LacticAcid'] = 0.85/0.86
    BD_dct['Lignin'] = BD_dct['SolubleLignin'] = 0.001/0.86
    BD_dct['Tar'] = 0.
    BD_dct['Xylose'] = 0.8/0.86
    BD_dct['XyloseOligomer'] = BD_dct['Xylan'] = 0.75/0.86

    # Assume biodegradabilities based on glucose and xylose
    BD_dct['Galactose'] = BD_dct['Mannose'] = \
        BD_dct['Arabinose']/BD_dct['Xylose'] * BD_dct['Glucose']
    BD_dct['GalactoseOligomer'] = BD_dct['Galactan'] = \
        BD_dct['MannoseOligomer'] = BD_dct['Mannan'] = \
             BD_dct['Glucan']

    BD_dct['ArabinoseOligomer'] = BD_dct['Arabinan'] = BD_dct['Xylan']

    if kwargs: # other input biodegradabilities
        BD_dct.update(kwargs)

    return BD_dct


def compute_stream_COD(stream):
    r"""
    Compute the chemical oxygen demand (COD) of a given stream in kg-O2/m3
    by summing the COD of each chemical in the stream using:

    .. math::
        COD [\frac{kg}{m^3}] = mol_{chemical} [\frac{kmol}{m^3}] * \frac{g O_2}{mol chemical}
    """
    chems = stream.chemicals
    mol = stream.mol
    if stream.F_vol == 0: return 0
    iCOD = np.array([-get_COD_stoichiometry(i)['O2'] for i in chems])
    COD = (mol*iCOD).sum()*molecular_weight({'O': 2}) / stream.F_vol
    return COD


def get_digestion_rxns(stream, BD, X_biogas, X_growth, biomass_ID):
    biomass_MW = getattr(stream.chemicals, biomass_ID).MW
    chems = [i for i in stream.chemicals if i.ID!=biomass_ID]
    BD = 1. if not BD else BD
    if isinstance(BD, float):
        BD = dict.fromkeys([i.ID for i in chems], BD)

    if X_biogas+X_growth > 1:
        raise ValueError('Sum of `X_biogas`/`X_decomp` and `X_biogas` is '
                         f'{X_biogas+X_growth}, larger than 100%.')

    biogas_rxns = []
    growth_rxns = []
    for i in chems:
        X = BD.get(i.ID)
        if not X:
            continue # assume no entry means not biodegradable

        biogas_stoyk = get_BMP_stoichiometry(i)
        if not biogas_stoyk.get(i.ID): # no conversion of this chemical
            continue

        iX_biogas = X * X_biogas # the amount of chemical used for biogas production
        iX_growth = X * X_growth # the amount of chemical used for cell growth

        if iX_biogas: # do not check atomic balance as P will not be accounted for
            biogas_rxn = Rxn(reaction=biogas_stoyk, reactant=i.ID, X=iX_biogas,
                             check_atomic_balance=False)
            biogas_rxns.append(biogas_rxn)

        if iX_growth:
        # Cannot check atom balance since the substrate may not have the atom
            growth_rxn = Rxn(f'{i.ID} -> {i.MW/biomass_MW}{biomass_ID}',
                             reactant=i.ID, X=iX_growth,
                             check_atomic_balance=False)


            growth_rxns.append(growth_rxn)

    if len(biogas_rxns)+len(growth_rxns)>1:
        return PRxn(biogas_rxns+growth_rxns)

    return []


# %%

# =============================================================================
# Miscellaneous
# =============================================================================

def format_str(string):
    string = string.replace(' ', '_')
    string = string.replace('-', '_')
    return string


def get_split_dct(chemicals, **split):
    # Copied from the corn stover biorefinery,
    # which is based on the 2011 NREL report (Humbird et al.),
    # assume no insolubles go to permeate
    insolubles_dct = dict.fromkeys(default_insolubles, 0.)
    split_dct = dict(
        Water=0.1454,
        Glycerol=0.125,
        LacticAcid=0.145,
        SuccinicAcid=0.125,
        HNO3=0.1454,
        Denaturant=0.125,
        DAP=0.1454,
        AmmoniumAcetate=0.145,
        AmmoniumSulfate=0.1454,
        H2SO4=0.1454,
        NaNO3=0.1454,
        Oil=0.125,
        N2=0.1351,
        NH3=0.1579,
        O2=0.15,
        CO2=0.1364,
        Xylose=0.25,
        Sucrose=0.125,
        Mannose=0.125,
        Galactose=0.125,
        Arabinose=0.125,
        Extract=0.145,
        NaOH=0.1454,
        SolubleLignin=0.145,
        GlucoseOligomer=0.1429,
        GalactoseOligomer=0.1429,
        MannoseOligomer=0.1429,
        XyloseOligomer=0.1429,
        ArabinoseOligomer=0.1429,
        Xylitol=0.125,
        Cellobiose=0.125,
        Cellulase=0.145
        )
    split_dct.update(insolubles_dct)
    default_chemical_dict(split_dct, chemicals, 0.15, 0.125, 0) # 'g', 'l', 's'

    if split is not None:
        split_dct.update(split)

    remove_undefined_chemicals(split_dct, chemicals)

    return split_dct


def kph_to_tpd(stream):
    dry_mass = stream.F_mass - stream.imass['Water']
    factor = 0.026455471462185312 # auom('kg').conversion_factor('ton')/auom('hr').conversion_factor('day')
    return dry_mass*factor


# %%

# =============================================================================
# Related to techno-economic analysis (TEA)
# =============================================================================

# Renewable natural gas, RIN D3, average of 2015 (first year with year-round data)-2021
# https://www.epa.gov/fuels-registration-reporting-and-compliance-help/rin-trades-and-price-information
# D3 designation based on entry Q of the approved pathways
# https://www.epa.gov/renewable-fuel-standard-program/approved-pathways-renewable-fuel
RIN_per_gal = 1.843 # $/gal ethanol

# According to the statue:
# https://www.ecfr.gov/current/title-40/chapter-I/subchapter-C/part-80/subpart-M/section-80.1415
# https://www.ecfr.gov/current/title-40/chapter-I/subchapter-C/part-80/subpart-M/section-80.1426
# "(5) 77,000 Btu (lower heating value) of compressed natural gas (CNG) or liquefied natural gas (LNG) shall represent one gallon of renewable fuel with an equivalence value of 1.0."
# All heating values from the H2 tools, accessed 9/27/2022
# https://h2tools.org/hyarc/calculator-tools/lower-and-higher-heating-values-fuels
LHV_LNG = 48.62 # MJ/kg, liquefied natural gas
MJ_per_BTU = 0.001055
LHV_LNG /= MJ_per_BTU # Btu/kg
LHV_ethanol = 77000 # Btu/gal per the statue instead of the 76330 from the H2 tools website
LNG_to_ethanol = LHV_LNG/LHV_ethanol # gal ethanol/kg LNG
RIN_price = RIN_per_gal * LNG_to_ethanol # $/kg LNG

prices = {
    'bisulfite': 0.08, # $/L
    'citric_acid': 0.22, # $/L
    'naocl': 0.14, # $/L
    'RIN': RIN_price, # in addition to the natural gas price
    }

# =============================================================================
# # Related to life cycle assessment (LCA)
# =============================================================================
# 100-year global warming potential (GWP) in kg CO2-eq/kg dry material
# All ecoinvent entries are from v3.8, allocation at the point of substitution

# Ecoinvent, market for sodium hydrogen sulfite, GLO,
# converted to 38% solution
bisulfite_CF = 1.2871 * 0.38

# Ecoinvent, market for citric acid, GLO
citric_acid_CF = 5.9048

# Ecoinvent, market for sodium hypochlorite, without water, in 15% solution state, RoW,
# converted to 12.5 wt% solution (15 vol%)
naocl_CF = 2.4871 * 0.125

# GREET, North America, from shale and conventional recovery, average US
ng_CF = 0.3899 + 1/16.04246*44.0095

GWP_CFs = {
    'Bisulfite': bisulfite_CF,
    'CitricAcid': citric_acid_CF,
    'CH4': ng_CF,
    'H2SO4': 43.3831/1e3, # assumed to be for the solution
    'HCl': 1.9683, # in the US, assumed to be concentrated solution
    'NaOCl': naocl_CF,
    'NaOH': 2.0092,
    'NH3': 2.6355,
}


def add_CFs(stream_registry, unit_registry, stream_CF_dct):
    has_steam = False
    for ID, key_factor in stream_CF_dct.items():
        if ID == 'steam':
            has_steam = True
            continue
        stream = stream_registry.search(ID) if isinstance(ID, str) \
            else getattr(unit_registry.search(ID[0]), ID[1])[ID[2]]
        if stream: # some streams might only exist in exist/new systems
            try:
                iter(key_factor)
                if isinstance(key_factor, str): key_factor = (key_factor,)
            except:
                key_factor = (key_factor,)
            key, factor = (key_factor[0], 1.) if len(key_factor) == 1 else key_factor

            stream.characterization_factors['GWP'] = GWP_CFs[key]*factor
    bst.PowerUtility.set_CF('GWP', *GWP_CFs['Electricity'])
    if has_steam:
        steam = stream_registry.search('steam')
        MJ = steam.H / 1e3 # enthalpy in kJ/hr
        steam.characterization_factors['GWP'] = MJ/steam.F_mass * GWP_CFs['Steam'] 


# -*- coding: utf-8 -*-
"""
"""
import biosteam as bst
import thermosteam as tmo
import numpy as np
from numpy.testing import assert_allclose

def test_flash():
    import biosteam as bst
    import numpy as np
    from numpy.testing import assert_allclose
    bst.settings.set_thermo(['Water', 'Ethanol'], ideal=True)
    feed_liquid = bst.Stream('feed_liquid', Water=50, Ethanol=50, phase='l', T = 400)
    feed_gas = bst.Stream('feed_gas', Water=50, Ethanol=50, phase='g', T=400)
    flash_1 = bst.StageEquilibrium('flash_1', ins=[feed_liquid, feed_gas], Q=0, P=101325, phases=('g', 'l'))
    flashes = [flash_1]
    for flash in flashes:
        flash.T = 350
        flash.K = np.array([0.5, 1.5])
        flash.B = 1.
        flash.outs[0].mol[:] = [50, 50]
        flash.outs[1].mol[:] = [50, 50]
    sys = bst.System('sys', path=flashes, algorithm='phenomena-oriented')
    sys.track_convergence()
    for i in range(10):
        sys.run_phenomena()
    phenomena_graph = sys.get_phenomena_graph(raise_error=False)
    shape = phenomena_graph.variable_profiles.shape
    assert shape == (30, 8)
        

def test_2_stage_flash():
    import biosteam as bst
    import numpy as np
    from numpy.testing import assert_allclose
    bst.settings.set_thermo(['Water', 'Ethanol'], ideal=True)
    feed_liquid = bst.Stream('feed_liquid', Water=50, Ethanol=50, phase='l', T = 400)
    feed_gas = bst.Stream('feed_gas', Water=50, Ethanol=50, phase='g', T=400)
    flash_1 = bst.StageEquilibrium('flash_1', ins=[feed_liquid, feed_gas], Q=0, P=101325, phases=('g', 'l'))
    flash_2 = bst.StageEquilibrium('flash_2', ins=[flash_1-0, feed_liquid], B=1, P=101325, phases=('g', 'l'))
    flashes = [flash_1, flash_2]
    for flash in flashes:
        flash.T = 350
        flash.K = np.array([0.5, 1.5])
        flash.B = 1.
        flash.outs[0].mol[:] = [50, 50]
        flash.outs[1].mol[:] = [50, 50]
    sys = bst.System('sys', path=flashes, algorithm='phenomena-oriented')
    sys.track_convergence()
    for i in range(10):
        sys.run_phenomena()
    phenomena_graph = sys.get_phenomena_graph(raise_error=False)
    shape = phenomena_graph.variable_profiles.shape
    assert shape == (30, 15)

def test_column():
    import biosteam as bst
    import numpy as np
    from numpy.testing import assert_allclose
    bst.settings.set_thermo(['Water', 'Ethanol'], cache=True)
    feed = bst.Stream('feed', Ethanol=80, Water=100, T=80.215 + 273.15)
    D1 = bst.MESHDistillation('D1', N_stages=5, ins=[feed], feed_stages=[2],
        outs=['vapor', 'liquid'],
        reflux=0.673, boilup=2.57,
        LHK=('Ethanol', 'Water') )
    D1.set_flow_rates(D1.hot_start())
    sys = bst.System('sys', path=[D1])
    sys.track_convergence(True)
    for i in range(10):
        sys.run_phenomena()
    phenomena_graph = sys.get_phenomena_graph(raise_error=False)
    shape = phenomena_graph.variable_profiles.shape
    assert shape == (30, 38)
 
def test_system():
    import biosteam as bst
    import numpy as np
    from numpy.testing import assert_allclose
    bst.settings.set_thermo(['Water', 'AceticAcid', 'EthylAcetate'], cache=True)
    solvent_feed_ratio = 1
    @bst.SystemFactory
    def system(ins, outs):
        feed = bst.Stream('feed', AceticAcid=6660, Water=43600)
        solvent = bst.Stream('solvent', EthylAcetate=65000)
        recycle = bst.Stream('recycle')
        LE = bst.MultiStageEquilibrium(
            N_stages=6, ins=[feed, solvent, recycle],
            feed_stages=(0, -1, -1),
            phases=('L', 'l'),
            maxiter=200,
            use_cache=True,
        )
        # DAA = bst.MultiStageEquilibrium(N_stages=6, ins=[LE-0], feed_stages=[3],
        #     outs=['vapor', 'liquid'],
        #     stage_specifications={0: ('Reflux', 0.673), -1: ('Boilup', 2.57)},
        #     maxiter=200,
        #     phases=('g', 'l'),
        #     use_cache=True,
        # )
        DEA = bst.MultiStageEquilibrium(N_stages=6, ins=[LE-1], feed_stages=[3],
            outs=['vapor', 'liquid'],
            stage_specifications={0: ('Reflux', 0.673), -1: ('Boilup', 2.57)},
            phases=('g', 'l'),
            maxiter=200,
            use_cache=True,
        )
        HX = bst.SinglePhaseStage(ins=DEA-0, outs=recycle, T=320, phase='l')
        
        chemicals = bst.settings.chemicals
        
        @LE.add_specification(run=True)
        def fresh_solvent_flow_rate():
            broth = feed.F_mass
            EtAc_recycle = recycle.imass['EthylAcetate']
            EtAc_required = broth * solvent_feed_ratio
            if EtAc_required < EtAc_recycle:
                recycle.F_mass *= EtAc_required / EtAc_recycle
                EtAc_recycle = recycle.imass['EthylAcetate']
            EtAc_fresh = EtAc_required - EtAc_recycle
            solvent.imass['EthylAcetate'] = max(
                0, EtAc_fresh
            )
        
        @solvent.material_balance
        def fresh_solvent_flow_rate():
            s = np.ones(chemicals.size)
            r = np.zeros(chemicals.size)
            v = r.copy()
            index = chemicals.index('EthylAcetate')
            r[index] = 1
            v[index] = solvent_feed_ratio * feed.F_mass / chemicals.EthylAcetate.MW
            return (
                {solvent: s,
                 recycle: r},
                 v
            )
    sys = system(algorithm='phenomena oriented', 
                    molar_tolerance=1e-9,
                    relative_molar_tolerance=1e-9,
                    maxiter=2000,
                    method='fixed-point')
    sys.track_convergence(True)
    for i in range(10):
        sys.run_phenomena()
    phenomena_graph = sys.get_phenomena_graph(raise_error=False)
    shape = phenomena_graph.variable_profiles.shape
    assert shape == (37, 133)
 
if __name__ == '__main__':
    test_flash()
    test_2_stage_flash()
    test_column()
    test_system()


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.

__all__ = ('Segment',)

class Segment:
    __slots__ = ('sink', 'sink_index', 'source', 'source_index')
    
    def __init__(self, start_stream, end_stream):
        self.sink = sink = start_stream.sink # Segment starts here
        self.sink_index = sink.ins.index(start_stream)
        self.source = source = end_stream.source # Segment ends here
        self.source_index = source.outs.index(end_stream)
        
    def disconnect(self, join_ends=False):
        end_stream = self.source.outs[self.source_index] # Segment includes end stream
        end_unit = end_stream.sink # Segment does not include end unit
        inlet_index = end_unit.ins.index(end_stream)
        end_unit.ins[inlet_index] = None
        if join_ends: 
            # Join the upstream and downstream ends.
            # This maintains the path but without the segment.
            end_unit.ins[inlet_index] = self.sink.ins[self.sink_index]
        self.sink.ins[self.sink_index] = None
            
    def insert(self, stream):
        stream.sink.ins.replace(stream, self.source.outs[self.source_index])
        self.sink.ins[self.sink_index] = stream
        
        



# -*- coding: utf-8 -*-
"""
.. contents:: :local:

.. autoclass:: biosteam.units.aerated_bioreactor.AeratedBioreactor
.. autoclass:: biosteam.units.aerated_bioreactor.GasFedBioreactor

References
----------
.. [1] Benz, G. T. Optimize Power Consumption in Aerobic Fermenters. 
    Chem. Eng. Progress 2003, 99 (5), 100–103.

.. [2] Benz, G. T. Bioreactor Design for Chemical Engineers. Chem. Eng.\
    Progress 2011, 21–26.

.. [3] Seider, W. D., Lewin,  D. R., Seader, J. D., Widagdo, S., Gani, R.,
    & Ng, M. K. (2017). Product and Process Design Principles. Wiley.

"""
import biosteam as bst
from .stirred_tank_reactor import AbstractStirredTankReactor
from math import pi
import numpy as np
from scipy.constants import g
import flexsolve as flx
from warnings import filterwarnings, catch_warnings
from scipy.optimize import minimize_scalar, minimize, least_squares, differential_evolution
from biosteam.units.design_tools import aeration

__all__ = (
    'AeratedBioreactor', 'ABR',
    'GasFedBioreactor', 'GFB',
)

class AeratedBioreactor(AbstractStirredTankReactor):
    """
    Same as StirredTankReactor but includes aeration. The agitator power may
    vary to minimize the total power requirement of both the compressor and agitator
    yet achieve the required oxygen transfer rate.
    
    Examples
    --------
    >>> import biosteam as bst
    >>> from biorefineries.sugarcane import chemicals
    >>> bst.settings.set_thermo(chemicals)
    >>> feed = bst.Stream('feed',
    ...                   Water=1.20e+05,
    ...                   Glucose=2.5e+04,
    ...                   units='kg/hr',
    ...                   T=32+273.15)
    >>> # Model oxygen uptake as combustion
    >>> rxn = bst.Rxn('Glucose + O2 -> H2O + CO2', reactant='Glucose', X=0.5, correct_atomic_balance=True) 
    >>> R1 = bst.AeratedBioreactor(
    ...     'R1', ins=[feed, bst.Stream('air', phase='g')], outs=('vent', 'product'), tau=12, V_max=500,
    ...     reactions=rxn,
    ... )
    >>> R1.simulate()
    >>> R1.show()
    AeratedBioreactor: R1
    ins...
    [0] feed  
        phase: 'l', T: 305.15 K, P: 101325 Pa
        flow (kmol/hr): Water    6.66e+03
                        Glucose  139
    [1] air  
        phase: 'g', T: 305.15 K, P: 101325 Pa
        flow (kmol/hr): O2  1.02e+03
                        N2  3.85e+03
    outs...
    [0] vent  
        phase: 'g', T: 305.15 K, P: 101325 Pa
        flow (kmol/hr): Water  135
                        CO2    416
                        O2     606
                        N2     3.85e+03
    [1] product  
        phase: 'l', T: 305.15 K, P: 101325 Pa
        flow (kmol/hr): Water    6.94e+03
                        Glucose  69.4
    
    """
    _N_ins = 2
    _N_outs = 2
    _ins_size_is_fixed = False
    auxiliary_unit_names = (
        'compressor',
        'air_cooler',
        *AbstractStirredTankReactor.auxiliary_unit_names
    )
    T_default = 273.15 + 32 
    P_default = 101325
    kW_per_m3_default = 0.2955 # Reaction in homogeneous liquid; reference [1]
    batch_default = True
    default_methods = {
        'Stirred tank': 'Riet',
        'Bubble column': 'Dewes',
    }
    def _init(
            self, reactions, theta_O2=0.5, Q_O2_consumption=None,
            optimize_power=None, design=None, method=None, kLa_kwargs=None,
            cooler_pressure_drop=None, compressor_isentropic_efficiency=None,
            **kwargs,
        ):
        if compressor_isentropic_efficiency is None: compressor_isentropic_efficiency = 0.85
        #: Isentropic efficiency of the compressor. Defaults to 0.85.
        self.compressor_isentropic_efficiency = compressor_isentropic_efficiency 
        #: Pressure drop at the cooler [Pa]. Defaults to 20684.28 Pa, a heuristic value for a gas.
        self.cooler_pressure_drop = 20684.28 if cooler_pressure_drop is None else cooler_pressure_drop
        AbstractStirredTankReactor._init(self, **kwargs)
        self.reactions = reactions
        self.theta_O2 = theta_O2 # Average concentration of O2 in the liquid as a fraction of saturation.
        self.Q_O2_consumption = Q_O2_consumption # Forced duty per O2 consummed [kJ/kmol].
        self.optimize_power = True if optimize_power is None else optimize_power
        if design is None: 
            design = 'Stirred tank'
        elif design not in aeration.kLa_method_names:
            raise ValueError(
                f"{design!r} is not a valid design; only "
                f"{list(aeration.kLa_method_names)} are valid"
            )
        self.design = design
        if method is None:
            method = self.default_methods[design]
        if (key:=(design, method)) in aeration.kLa_methods:
            self.kLa = aeration.kLa_methods[key]
        elif hasattr(method, '__call__'):
            self.kLa = method
        else:
            raise ValueError(
                f"{method!r} is not a valid kLa method; only "
                f"{aeration.kLa_method_names[design]} are valid"
            )
        self.kLa_kwargs = {} if kLa_kwargs is None else kLa_kwargs
    
    def get_kLa(self):
        if self.kLa is aeration.kLa_stirred_Riet:
            V = self.get_design_result('Reactor volume', 'm3') * self.V_wf
            operating_time = self.tau / self.design_results.get('Batch time', 1.)
            N_reactors = self.parallel['self']
            P = 1000 * self.kW_per_m3 * V # W
            air_in = self.sparged_gas
            D = self.get_design_result('Diameter', 'm')
            F = air_in.get_total_flow('m3/s') / N_reactors / operating_time
            R = 0.5 * D
            A = pi * R * R
            self.superficial_gas_flow = U = F / A # m / s 
            return aeration.kLa_stirred_Riet(P, V, U, **self.kLa_kwargs) # 1 / s 
        elif self.kLa is aeration.kla_bubcol_Dewes:
            V = self.get_design_result('Reactor volume', 'm3') * self.V_wf
            operating_time = self.tau / self.design_results.get('Batch time', 1.)
            N_reactors = self.parallel['self']
            air_in = self.sparged_gas
            D = self.get_design_result('Diameter', 'm')
            F = air_in.get_total_flow('m3/s') / N_reactors / operating_time
            R = 0.5 * D
            A = pi * R * R
            self.superficial_gas_flow = U = F / A # m / s 
            feed = self.ins[0]
            try:
                return aeration.kla_bubcol_Dewes(U, feed.get_property('mu', 'mPa*s'), air_in.get_property('rho', 'kg/m3'))
            except:
                breakpoint()
        else:
            raise NotImplementedError('kLa method has not been implemented in BioSTEAM yet')
    
    def get_agitation_power(self, kLa):
        if self.kLa is aeration.kLa_stirred_Riet:
            air_in = self.sparged_gas
            N_reactors = self.parallel['self']
            operating_time = self.tau / self.design_results.get('Batch time', 1.)
            V = self.get_design_result('Reactor volume', 'm3') * self.V_wf
            D = self.get_design_result('Diameter', 'm')
            F = air_in.get_total_flow('m3/s') / N_reactors / operating_time
            R = 0.5 * D
            A = pi * R * R
            self.superficial_gas_flow = U = F / A # m / s
            return aeration.P_at_kLa_Riet(kLa, V, U, **self.kLa_kwargs)
        else:
            raise NotImplementedError('kLa method has not been implemented in BioSTEAM yet')
    
    def _get_duty(self):
        if self.Q_O2_consumption is None:
            H_in = sum([i.H for i in self.ins if i.phase != 'g'], self.air_cooler.outs[0].H)
            return self.H_out - H_in + self.Hf_out - self.Hf_in
        else:
            return self.Q_O2_consumption * (
                sum([i.imol['O2'] for i in self.ins])
                - sum([i.imol['O2'] for i in self.outs])
            )
    
    @property
    def feed(self):
        return self._ins[0]
    
    @property
    def air(self):
        for i in self._ins:
            if i.phase == 'g': return i
    
    @property
    def sparged_gas(self):
        return self.air_cooler.outs[0]
    
    @property
    def vent(self):
        return self._outs[0]
    
    def load_auxiliaries(self):
        super().load_auxiliaries()
        compressor = self.auxiliary(
            'compressor', bst.IsentropicCompressor, 
            self.air, eta=self.compressor_isentropic_efficiency, P=2 * 101325,
        )
        self.auxiliary(
            'air_cooler', bst.HXutility, compressor-0, T=self.T, 
            dP=False, 
        )
        
    def _run_vent(self, vent, effluent):
        vent.receive_vent(effluent, energy_balance=False, ideal=True)
        
    def _run(self):
        air = self.air
        if air is None:
            air = bst.Stream(phase='g', thermo=self.thermo)
            self.ins.insert(1, air)
            self.compressor.ins[0] = self.auxlet(air)
        feeds = [i for i in self.ins if i.phase != 'g']
        vent, effluent = self.outs
        air.P = vent.P = effluent.P = self.P
        air.T = vent.T = effluent.T = self.T
        vent.empty()
        vent.phase = 'g'
        air.phase = 'g'
        air.empty()
        compressor = self.compressor
        effluent.mix_from(feeds, energy_balance=False)
        self._run_reactions(effluent)
        effluent_no_air_data = effluent.get_data()
        OUR = -effluent.get_flow('mol/s', 'O2') # Oxygen uptake rate
        if OUR <= 1e-2:
            if OUR > 0: effluent.imol['O2'] = 0.
            self._run_vent(vent, effluent)
            return
        air_cc = self.sparged_gas
        air_cc.copy_like(air)
        air_cc.P = compressor.P = self._inlet_air_pressure()
        air_cc.T = self.T
        
        if self.optimize_power:
            def total_power_at_oxygen_flow(O2):
                air.set_flow([O2, O2 * 79. / 21.], 'mol/s', ['O2', 'N2'])
                air_cc.copy_flow(air) # Skip simulation of air cooler
                compressor.simulate()
                effluent.set_data(effluent_no_air_data)
                effluent.mix_from([effluent, air_cc], energy_balance=False)
                vent.empty()
                self._run_vent(vent, effluent)
                total_power = self._solve_total_power(OUR)
                return total_power
            
            f = total_power_at_oxygen_flow
            minimize_scalar(f, 1.2 * OUR, bounds=[OUR, 10 * OUR], tol=OUR * 1e-3)
        else:
            def air_flow_rate_objective(O2):
                air.set_flow([O2, O2 * 79. / 21.], 'mol/s', ['O2', 'N2'])
                air_cc.copy_flow(air) # Skip simulation of air cooler
                compressor.simulate()
                effluent.set_data(effluent_no_air_data)
                effluent.mix_from([effluent, air_cc], energy_balance=False)
                vent.empty()
                self._run_vent(vent, effluent)
                return OUR - self.get_OTR()
            
            f = air_flow_rate_objective
            y0 = air_flow_rate_objective(OUR)
            if y0 <= 0.: # Correlation is not perfect and special cases lead to OTR > OUR
                return
            flx.IQ_interpolation(f, x0=OUR, x1=10 * OUR, 
                                 y0=y0, ytol=1e-3, xtol=1e-3)
        
    def _run_reactions(self, effluent):
        self.reactions.force_reaction(effluent)
    
    def _solve_total_power(self, OUR): # For OTR = OUR [mol / s]
        air_in = self.sparged_gas
        N_reactors = self.parallel['self']
        operating_time = self.tau / self.design_results.get('Batch time', 1.)
        V = self.get_design_result('Reactor volume', 'm3') * self.V_wf
        vent = self.vent
        P_O2_air = air_in.get_property('P', 'bar') * air_in.imol['O2'] / air_in.F_mol
        P_O2_vent = 0. if vent.isempty() else vent.get_property('P', 'bar') * vent.imol['O2'] / vent.F_mol
        C_O2_sat_air = aeration.C_O2_L(self.T, P_O2_air) # mol / kg
        C_O2_sat_vent = aeration.C_O2_L(self.T, P_O2_vent) # mol / kg
        theta_O2 = self.theta_O2
        LMDF = aeration.log_mean_driving_force(C_O2_sat_vent, C_O2_sat_air, theta_O2 * C_O2_sat_vent, theta_O2 * C_O2_sat_air)
        kLa = OUR / (LMDF * V * self.effluent_density * N_reactors * operating_time) 
        P = self.get_agitation_power(kLa)
        agitation_power_kW = P / 1000
        total_power_kW = (agitation_power_kW + self.compressor.power_utility.consumption / N_reactors) / V
        self.kW_per_m3 = agitation_power_kW / V 
        return total_power_kW
    
    def get_OUR(self):
        """Return the oxygen uptake rate in mol/s."""
        feeds = [i for i in self.ins if i.phase != 'g']
        effluent = self.effluent.copy()
        effluent.mix_from(feeds, energy_balance=False)
        self._run_reactions(effluent)
        return -effluent.get_flow('mol/s', 'O2') # Oxygen uptake rate
    
    def get_OTR(self):
        """Return the oxygen transfer rate in mol/s."""
        V = self.get_design_result('Reactor volume', 'm3') * self.V_wf
        operating_time = self.tau / self.design_results.get('Batch time', 1.)
        N_reactors = self.parallel['self']
        kLa = self.get_kLa()
        air_in = self.sparged_gas
        vent = self.vent
        P_O2_air = air_in.get_property('P', 'bar') * air_in.imol['O2'] / air_in.F_mol
        P_O2_vent = vent.get_property('P', 'bar') * vent.imol['O2'] / vent.F_mol
        C_O2_sat_air = aeration.C_O2_L(self.T, P_O2_air) # mol / kg
        C_O2_sat_vent = aeration.C_O2_L(self.T, P_O2_vent) # mol / kg
        theta_O2 = self.theta_O2
        LMDF = aeration.log_mean_driving_force(C_O2_sat_vent, C_O2_sat_air, theta_O2 * C_O2_sat_vent, theta_O2 * C_O2_sat_air)
        OTR = kLa * LMDF * self.effluent_density * V * N_reactors * operating_time # mol / s
        return OTR
        
    def _inlet_air_pressure(self):
        AbstractStirredTankReactor._design(self)
        liquid = bst.Stream(None, thermo=self.thermo)
        liquid.mix_from([i for i in self.ins if i.phase != 'g'], energy_balance=False)
        liquid.copy_thermal_condition(self.outs[0])
        self.effluent_density = rho = liquid.rho
        length = self.get_design_result('Length', 'm') * self.V_wf
        return g * rho * length + 101325 + self.cooler_pressure_drop # Pa
    
    def _design(self):
        AbstractStirredTankReactor._design(self)
        if self.air.isempty(): return
        liquid = bst.Stream(None, thermo=self.thermo)
        liquid.mix_from([i for i in self.ins if i.phase != 'g'], energy_balance=False)
        liquid.copy_thermal_condition(self.outs[0])
        rho = liquid.rho
        length = self.get_design_result('Length', 'm') * self.V_wf
        compressor = self.compressor
        compressor.P = g * rho * length + 101325
        compressor.simulate()
        air_cooler = self.air_cooler
        air_cooler.T = self.T
        air_cooler.simulate()
        self.parallel['compressor'] = 1
        self.parallel['air_cooler'] = 1
        # For robust process control, do not include in HXN
        for unit in self.auxiliary_units:
            for hu in unit.heat_utilities: hu.hxn_ok = False


class GasFedBioreactor(AbstractStirredTankReactor):
    """
    Same as AbstractStirredTankReactor but includes multiple gas feeds. The agitator power may
    vary to minimize the total power requirement of both the compressor and agitator
    yet achieve the required oxygen transfer rate.
    
    Examples
    --------
    >>> import biosteam as bst
    >>> bst.settings.set_thermo(['H2', 'CO2', 'N2', 'O2', 'H2O', 'AceticAcid'])
    >>> media = bst.Stream(ID='media', H2O=10000, units='kg/hr')
    >>> H2 = bst.Stream(ID='H2', H2=100, units='kg/hr', phase='g')
    >>> fluegas = bst.Stream(ID='fluegas', N2=70, CO2=25, H2O=3, O2=2, units='kg/hr', phase='g')
    >>> # Model acetic acid production from H2 and CO2
    >>> rxn = bst.Rxn('H2 + CO2 -> AceticAcid + H2O', reactant='H2', correct_atomic_balance=True) 
    >>> brxn = rxn.backwards(reactant='AceticAcid')
    >>> R1 = bst.GasFedBioreactor(
    ...     'R1', ins=[media, H2, fluegas], outs=('vent', 'product'), tau=68, V_max=500,
    ...     reactions=rxn, backward_reactions=brxn,
    ...     gas_substrates=('H2', 'CO2'),
    ...     titer={'AceticAcid': 5},
    ...     optimize_power=False,
    ...     kW_per_m3=0.,
    ... )
    >>> R1.simulate()
    >>> R1.show()
    GasFedBioreactor: R1
    ins...
    [0] media  
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): H2O  188
    [1] H2  
        phase: 'g', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): H2  49.6
    [2] fluegas  
        phase: 'g', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): CO2  0.568
                        N2   2.5
                        O2   0.0625
                        H2O  0.167
    outs...
    [0] vent  
        phase: 'g', T: 305.15 K, P: 101325 Pa
        flow (kmol/hr): H2          48.5
                        N2          2.5
                        O2          0.0625
                        H2O         1.89
                        AceticAcid  0.00183
    [1] product  
        phase: 'l', T: 305.15 K, P: 101325 Pa
        flow (kmol/hr): H2O         187
                        AceticAcid  0.282
    
    """
    _N_ins = 2
    _N_outs = 2
    _ins_size_is_fixed = False
    auxiliary_unit_names = (
        'sparger',
        'compressors',
        'gas_coolers',
        *AbstractStirredTankReactor.auxiliary_unit_names
    )
    T_default = 273.15 + 32 
    P_default = 101325
    kW_per_m3_default = 0.2955 # Reaction in homogeneous liquid
    batch_default = True
    default_methods = AeratedBioreactor.default_methods
    get_kLa = AeratedBioreactor.get_kLa
    get_agitation_power = AeratedBioreactor.get_agitation_power
    
    def _init(self, 
            reactions, gas_substrates, 
            # Can vary either liquid or gas flows
            titer=None, 
            # Only for variable gas flows
            backward_reactions=None, 
            variable_gas_feeds=(), 
            # General design/performance arguments
            design=None, method=None, kLa_kwargs=None,
            theta=0.5, Q_consumption=None,
            cooler_pressure_drop=None,
            # Only for agitated bioreactors (not bubble column)
            optimize_power=None, 
            **kwargs,
        ):
        self.cooler_pressure_drop = 20684.28 if cooler_pressure_drop is None else cooler_pressure_drop
        self.reactions = reactions
        self.backward_reactions = backward_reactions
        self.theta = theta # Average concentration of gas substrate in the liquid as a fraction of saturation.
        self.Q_consumption = Q_consumption # Forced duty per gas substrate consummed [kJ/kmol].
        self.kLa_kwargs = {} if kLa_kwargs is None else kLa_kwargs
        self.optimize_power = True if optimize_power is None else optimize_power
        self.variable_gas_feeds = variable_gas_feeds # list[int|Stream] Feed index or stream.
        self.gas_substrates = gas_substrates
        self.titer = titer # dict[str, float] g / L
        AbstractStirredTankReactor._init(self, **kwargs)
        if design is None: 
            if self.kW_per_m3 == 0:
                design = 'Bubble column'
            else:
                design = 'Stirred tank'
        elif design not in aeration.kLa_method_names:
            raise ValueError(
                f"{design!r} is not a valid design; only "
                f"{list(aeration.kLa_method_names)} are valid"
            )
        self.design = design
        if method is None:
            method = self.default_methods[design]
        if (key:=(design, method)) in aeration.kLa_methods:
            self.kLa = aeration.kLa_methods[key]
        elif hasattr(method, '__call__'):
            self.kLa = method
        else:
            raise ValueError(
                f"{method!r} is not a valid kLa method; only "
                f"{aeration.kLa_method_names[design]} are valid"
            )
    
    def _get_duty(self):
        if self.Q_consumption is None:
            H_in = sum(
                [i.H for i in self.ins if i.phase != 'g']
                + [i.outs[0].H for i in self.gas_coolers]
            )
            return self.H_out - H_in + self.Hf_out - self.Hf_in
        else:
            return self.Q_consumption * (
                sum([i.imol['O2'] for i in self.ins])
                - sum([i.imol['O2'] for i in self.outs])
            )
    
    @property
    def vent(self):
        return self._outs[0]
    
    @property
    def variable_gas_feeds(self):
        return [(i if isinstance(i, bst.Stream) else self.ins[i]) for i in self._variable_gas_feeds]
    @variable_gas_feeds.setter
    def variable_gas_feeds(self, variable_gas_feeds):
        self._variable_gas_feeds = variable_gas_feeds
    
    @property
    def normal_gas_feeds(self):
        variable = set(self.variable_gas_feeds)
        return [i for i in self.ins if i not in variable and i.phase == 'g']
    
    @property
    def liquid_feeds(self):
        return [i for i in self.ins if i.phase != 'g']
    
    @property
    def sparged_gas(self):
        return self.sparger-0
    
    def load_auxiliaries(self):
        super().load_auxiliaries()
        self.compressors = []
        self.gas_coolers = []
        for i in self.variable_gas_feeds: i.phase = 'g'
        for inlet in self.ins:
            if inlet.phase != 'g': continue
            compressor = self.auxiliary(
                'compressors', bst.IsentropicCompressor, inlet, eta=0.85, P=2 * 101325
            )
            self.auxiliary(
                'gas_coolers', bst.HXutility, compressor-0, T=self.T
            )
        self.auxiliary(
            'sparger', bst.Mixer, [i-0 for i in self.gas_coolers]
        )
        
    def _run_vent(self, vent, effluent):
        vent.receive_vent(effluent, energy_balance=False, ideal=True)
        
    def get_SURs(self, effluent):
        F_vol = effluent.ivol['Water'] # m3 / hr
        produced = bst.Stream(None, thermo=self.thermo)
        for ID, concentration in self.titer.items():
            produced.imass[ID] = F_vol * concentration
        consumed = produced.copy()
        self.backward_reactions.force_reaction(consumed)
        return consumed.get_flow('mol/s', self.gas_substrates), consumed, produced
    
    def _load_gas_feeds(self):
        for i in self.compressors: i.simulate()
        for i in self.gas_coolers: i.simulate()
        self.sparger.simulate()
    
    def _run(self):
        variable_gas_feeds = self.variable_gas_feeds
        vent, effluent = self.outs
        vent.P = effluent.P = self.P
        vent.T = effluent.T = self.T
        vent.empty()
        vent.phase = 'g'
        if not self.titer:
            liquid_feeds = [i for i in self.ins if i.phase != 'g']
            T = self.T
            P = self._inlet_gas_pressure()
            for i in self.compressors: i.P = P
            for i in self.gas_coolers: i.T = T
            self._load_gas_feeds()
            STRs = self.get_STRs()
            effluent.mix_from(liquid_feeds, energy_balance=False)
            effluent.set_flow(STRs, units='mol/s', key=self.gas_substrates)
            self._run_reactions(effluent)
            vent.empty()
            self._run_vent(vent, effluent)
        elif variable_gas_feeds:
            liquid_feeds = [i for i in self.ins if i.phase != 'g']
            effluent.mix_from(liquid_feeds, energy_balance=False)
            T = self.T
            P = self._inlet_gas_pressure()
            for i in self.compressors: i.P = P
            for i in self.gas_coolers: i.T = T
            effluent_liquid_data = effluent.get_data()
            SURs, s_consumed, s_produced = self.get_SURs(effluent) # Gas substrate uptake rate [mol / s]
            if (SURs <= 1e-2).all():
                effluent.imol[self.gas_substrates] = 0.
                self._run_vent(vent, effluent)
                return
            x_substrates = []
            for gas, ID in zip(self.variable_gas_feeds, self.gas_substrates):
                x_substrates.append(gas.get_molar_fraction(ID))
            index = range(len(self.gas_substrates))
            
            def load_flow_rates(F_feeds):
                for i in index:
                    gas = variable_gas_feeds[i]
                    gas.set_total_flow(F_feeds[i], 'mol/s')
                self._load_gas_feeds()
                effluent.set_data(effluent_liquid_data)
                effluent.mix_from([self.sparged_gas, -s_consumed, s_produced, *liquid_feeds], energy_balance=False)
                if (effluent.mol < 0).any(): breakpoint()
                vent.empty()
                self._run_vent(vent, effluent)
            
            baseline_feed = bst.Stream.sum(self.normal_gas_feeds, energy_balance=False)
            baseline_flows = baseline_feed.get_flow('mol/s', self.gas_substrates)
            bounds = np.array([[max(1.01 * SURs[i] - baseline_flows[i], 1e-6), 10 * SURs[i]] for i in index])
            if self.optimize_power:
                def total_power_at_substrate_flow(F_substrates):
                    load_flow_rates(F_substrates)
                    total_power = self._solve_total_power(SURs)
                    return total_power
                
                f = total_power_at_substrate_flow
                with catch_warnings():
                    filterwarnings('ignore')
                    results = minimize(f, 1.2 * SURs, bounds=bounds, tol=SURs.max() * 1e-6)
                    load_flow_rates(results.x / x_substrates)
            else:
                def gas_flow_rate_objective(F_substrates):
                    F_feeds = F_substrates / x_substrates
                    load_flow_rates(F_feeds)
                    STRs = self.get_STRs() # Must meet all substrate demands
                    F_ins = F_substrates + baseline_flows
                    mask = STRs - F_ins > 0
                    STRs[mask] = F_ins[mask]
                    diff = SURs - STRs
                    diff[diff > 0] *= 1e3 # Force transfer rate to meet uptake rate
                    SE = (diff * diff).sum()
                    return SE
                
                f = gas_flow_rate_objective
                with catch_warnings():
                    filterwarnings('ignore')
                    bounds = bounds.T
                    results = least_squares(f, 1.2 * SURs, bounds=bounds, ftol=SURs.min() * 1e-6)
                self._results = results
                load_flow_rates(results.x / x_substrates)
        else:
            try:
                feed, = [i for i in self.ins if i.phase != 'g']
            except:
                raise RuntimeError('gas-fed bioreactor must have exactly on liquid feed')
            T = self.T
            P = self._inlet_gas_pressure()
            for i in self.compressors: i.P = P
            for i in self.gas_coolers: i.T = T
            self._load_gas_feeds()
            F_liquid_max = self._initialize_variable_liquid_guess(feed, effluent, vent)
            product, titer = next(iter(self.titer.items()))
            def liquid_flow_rate_objective(F_feed):
                feed.F_mass = F_feed
                STRs = self.get_STRs()
                effluent.copy_flow(feed)
                effluent.set_flow(STRs, units='mol/s', key=self.gas_substrates)
                self._run_reactions(effluent)
                vent.empty()
                self._run_vent(vent, effluent)
                return effluent.imass[product] / effluent.ivol['Water'] - titer
                
            flx.IQ_interpolation(liquid_flow_rate_objective, 0.05 * F_liquid_max, F_liquid_max, ytol=1e-3)
        # self.show()
        # breakpoint()
        
    def _run_reactions(self, effluent):
        effluent.mix_from(self.ins)
        data = effluent.get_data()
        rxns = self.reactions
        rxns.force_reaction(effluent)
        for reactant in self.gas_substrates:
            if effluent.imol[reactant] < 0:
                if isinstance(rxns, bst.Rxn):
                    rxns.reactant = reactant
                else:
                    for rxn in rxns:
                        if rxn.istoichiometry[reactant] < 0:
                            rxn.reactant = reactant
            
                effluent.set_data(data)
                rxns.force_reaction(effluent)
                break
        
    def _initialize_variable_liquid_guess(self, feed, effluent, vent):
        effluent.copy_flow(feed)
        self._run_reactions(effluent)
        product, titer = next(iter(self.titer.items()))
        F_liquid_max = 1000 * effluent.imass[product] / titer # kg / hr
        feed.F_mass = F_liquid_max
        effluent.mix_from(self.ins, energy_balance=False)
        vent.empty()
        self._run_vent(vent, effluent)
        return F_liquid_max
        
    def _solve_total_power(self, SURs): # For STR = SUR [mol / s]
        gas_in = self.sparged_gas
        N_reactors = self.parallel['self']
        operating_time = self.tau / self.design_results.get('Batch time', 1.)
        V = self.get_design_result('Reactor volume', 'm3') * self.V_wf
        D = self.get_design_result('Diameter', 'm')
        F = gas_in.get_total_flow('m3/s') / N_reactors / operating_time
        R = 0.5 * D
        A = pi * R * R
        self.superficial_gas_flow = U = F / A # m / s 
        vent = self.vent
        Ps = []
        for gas_substrate, SUR in zip(self.gas_substrates, SURs):
            Py_gas = gas_in.get_property('P', 'bar') * gas_in.imol[gas_substrate] / gas_in.F_mol
            Py_vent = 0. if vent.isempty() else vent.get_property('P', 'bar') * vent.imol[gas_substrate] / vent.F_mol
            C_sat_gas = aeration.C_L(self.T, Py_gas, gas_substrate) # mol / kg
            C_sat_vent = aeration.C_L(self.T, Py_vent, gas_substrate) # mol / kg
            theta = self.theta
            LMDF = aeration.log_mean_driving_force(C_sat_vent, C_sat_gas, theta * C_sat_vent, theta * C_sat_gas)
            kLa = SUR / (LMDF * V * self.effluent_density * N_reactors * operating_time)
            Ps.append(aeration.P_at_kLa_Riet(kLa, V, U, **self.kLa_kwargs))
        P = max(Ps)  
        agitation_power_kW = P / 1000
        compressor_power_kW = sum([i.power_utility.consumption for i in self.compressors]) / N_reactors
        total_power_kW = (agitation_power_kW + compressor_power_kW) / V
        self.kW_per_m3 = agitation_power_kW / V 
        return total_power_kW
    
    def get_STRs(self):
        """Return the gas substrate transfer rate in mol/s."""
        V = self.get_design_result('Reactor volume', 'm3') * self.V_wf
        operating_time = self.tau / self.design_results.get('Batch time', 1.)
        N_reactors = self.parallel['self']
        gas_in = self.sparged_gas
        kLa = self.get_kLa() # 1 / s 
        vent = self.vent
        P_gas = gas_in.get_property('P', 'bar')
        P_vent = vent.get_property('P', 'bar')
        STRs = []
        for ID in self.gas_substrates:
            Py_gas = P_gas * gas_in.imol[ID] / gas_in.F_mol
            Py_vent = P_vent * vent.imol[ID] / vent.F_mol
            C_sat_gas = aeration.C_L(self.T, Py_gas, ID) # mol / kg
            C_sat_vent = aeration.C_L(self.T, Py_vent, ID) # mol / kg
            theta = self.theta
            LMDF = aeration.log_mean_driving_force(C_sat_vent, C_sat_gas, theta * C_sat_vent, theta * C_sat_gas)
            STRs.append(
                kLa * LMDF * self.effluent_density * V * N_reactors * operating_time # mol / s
            )
        return np.array(STRs)
        
    def _inlet_gas_pressure(self):
        AbstractStirredTankReactor._design(self, size_only=True)
        liquid = bst.Stream(None, thermo=self.thermo)
        liquid.mix_from([i for i in self.ins if i.phase != 'g'], energy_balance=False)
        liquid.copy_thermal_condition(self.outs[0])
        self.effluent_density = rho = liquid.rho
        length = self.get_design_result('Length', 'm') * self.V_wf
        return g * rho * length + 101325 + self.cooler_pressure_drop # Pa
    
    def _design(self):
        AbstractStirredTankReactor._design(self)
        self.parallel['sparger'] = 1
        self.parallel['mixers'] = 1
        self.parallel['compressors'] = 1
        self.parallel['gas_coolers'] = 1
        # For robust process control, do not include in HXN
        for unit in self.auxiliary_units:
            for hu in unit.heat_utilities: hu.hxn_ok = False
    
GFB = GasFedBioreactor
ABR = AeratedBioreactor



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
.. contents:: :local:

.. autodata:: stream_utility_prices
.. autodata:: CE

"""
from __future__ import annotations
__version__ = '2.51.15'

#: Chemical engineering plant cost index (defaults to 567.5 at 2017).
CE: float = 567.5 

#: User-defined impact indicators and their units of measure.
impact_indicators: dict[str, str] = {}

#: Price of stream utilities/fees/credits [USD/kg] which are defined as 
#: inlets and outlets to unit operations.
stream_prices: dict[str, float] = {}

stream_utility_prices = stream_prices # For backwards compatibility

#: Defined allocation property and basis pairs for LCA.
allocation_properties: dict[str, str] = {}

# %% Workaround for readthedocs, which fails to cache numba

import numba
try:
    @numba.njit(cache=True)
    def f_dummy(): pass
except RuntimeError: # pragma: no cover
    def njit(*args, **kwargs):
        kwargs['cache'] = False
        return numba.jit(*args, **kwargs)
    numba.njit = njit
    del njit
else:
    del f_dummy

# %% Initialize BioSTEAM 

import thermosteam
from thermosteam import *
import thermosteam.units_of_measure # Import the units_of_measure module to override the units_of_measure decorator.
from ._heat_utility import UtilityAgent, HeatUtility
from ._power_utility import PowerUtility
from . import plots
from .utils import *
from ._unit import Unit
from . import _system
from ._system import *
from . import _flowsheet
from ._flowsheet import *
from . import process_tools
from .process_tools import *
from . import _tea
from ._tea import *
from . import utils
from . import units
from .units import *
from ._facility import Facility
from . import _module
from ._module import *
from . import facilities
from .facilities import *
from . import wastewater
from .wastewater import *
from . import evaluation
from .evaluation import *
from . import exceptions
from . import report
from . import _settings

__all__ = (
    'Unit', 'PowerUtility', 'UtilityAgent', 'HeatUtility', 'Facility',
    'utils', 'units', 'facilities', 'wastewater', 'evaluation', 'Chemical', 'Chemicals', 'Stream',
    'MultiStream', 'settings', 'exceptions', 'report', 'units_of_measure',
    'process_tools', 'preferences', *_system.__all__, *_flowsheet.__all__, 
    *_tea.__all__, *units.__all__, *facilities.__all__, *wastewater.__all__,
    *evaluation.__all__, *process_tools.__all__, *_module.__all__,
)

def nbtutorial(dark=False):
    main_flowsheet.clear()
    preferences.reset()
    if dark: 
        preferences.dark_mode(bg='#111111')
    else:
        preferences.light_mode(bg='#ffffffaa')
    preferences.tooltips_full_results = False
    preferences.graphviz_format = 'html'
    preferences.show_all_streams = True
    from warnings import filterwarnings
    filterwarnings('ignore')

# %% 
import pandas as pd
from IPython.display import display

def display_table_as_html(series):
    return display(pd.DataFrame(series, columns=['']))

pd.Series._ipython_display_ = display_table_as_html


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import pytest
import biosteam as bst
import numpy as np
from thermosteam.network import Network
from numpy.testing import assert_allclose

def test_auxiliary_unit_owners():
    with bst.System() as sys:
        class NewUnit(bst.Unit):
            auxiliary_unit_names = ('mixer',)
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.mixer = bst.Mixer()
                
        bst.settings.set_thermo(['Water'], cache=True)
        unit = NewUnit()
    # Once unit is in a system, auxiliary units must have an owner
    assert unit.mixer.owner is unit

def test_unit_convinience_properties():
    class TestUnit(bst.Unit):
        _N_ins = 2
        _N_outs = 2
        
        def _run(self):
            for i, j in zip(self.ins, self.outs):
                j.copy_like(i)
    
    bst.settings.set_thermo(['Water', 'Ethanol'], cache=True)
    ins = [bst.Stream(None, Water=1, Ethanol=2),
           bst.Stream(None, Water=2, Ethanol=5)]
    U = TestUnit(None, ins=ins, outs=(None, None))
    U.run()
    assert (U.mol_in == [3, 7]).all()
    assert (U.mol_in == U.mol_out).all()
    assert (U.z_mol_in == U.z_mol_out).all()
    assert (U.z_mol_in == [0.3, 0.7]).all()

def test_unit_inheritance_setup_method():
    class NewUnit(bst.Unit):
        def _setup(self):
            super()._setup()
    
    bst.settings.set_thermo(['Water'], cache=True)
    U1 = NewUnit()
    U1.add_specification(lambda: None)
    U1.simulate()

def test_process_specifications_linear():
    bst.settings.set_thermo(['Water', 'Ethanol'], cache=True)
    with bst.System() as sys:
        T1 = bst.StorageTank('T1', ins=bst.Stream(Water=1000))
        H1 = bst.HXutility('H1', ins=T1-0, T=320)
        T2 = bst.StorageTank('T2', ins=bst.Stream(Ethanol=1000))
        H2 = bst.HXutility('H2', ins=T2-0, T=320)
        T3 = bst.StorageTank('T3', ins=bst.Stream(Water=10))
        H3 = bst.HXutility('H3', ins=T3-0, T=320)
        M1 = bst.Mixer('M1', ins=[H1-0, H2-0, H3-0])
        H4 = bst.HXutility('H4', ins=M1-0, T=350)
    
    # Specification impacting upstream units
    @M1.add_specification(run=True, impacted_units=[T1])
    def adjust_flow_rate():
        water = T1.ins[0]
        ethanol = T2.ins[0]
        water.F_mass = ethanol.F_mass
    sys.simulate()
    assert M1.specifications[0].path == [T1, H1]
    assert sys._to_network() == Network(
        [T2,
         H2,
         T1,
         H1,
         T3,
         H3,
         M1,
         H4]
    )
    assert_allclose(H4.outs[0].mol, sum([i.ins[0].mol for i in (T1, T2, T3)]))
    
    # Specification impacting units in parallel (neither upstream nor downstream units).
    # System simulation order must switch
    old_path_length = len(sys.unit_path)
    M1.specifications.clear() # Remove specification
    tanks = [i for i in sys.units if isinstance(i, bst.Tank)]
    first_tank, mid_tank, last_tank = tanks
    @last_tank.add_specification(run=True, impacted_units=[first_tank, mid_tank])
    def adjust_flow_rate():
        mid_tank.ins[0].F_vol = last_tank.ins[0].F_vol
        first_tank.ins[0].F_mol = 1.0
    
    @mid_tank.add_specification(run=True, impacted_units=[first_tank])
    def adjust_flow_rate():
        first_tank.ins[0].F_vol = mid_tank.ins[0].F_vol
    
    sys.simulate()
    # No upstream unit simulations
    assert mid_tank.specifications[0].path == []
    assert last_tank.specifications[0].path == []
    # Order of tank simulation is now reversed
    assert [last_tank, mid_tank, first_tank] == [i for i in sys.units if isinstance(i, bst.Tank)] 
    # 2 hidden connection source and 2 sinks
    assert len(sys.unit_path) == old_path_length
    assert (H4.outs[0].mol == sum([i.ins[0].mol for i in (T1, T2, T3)])).all()
    assert sys.path[0] is last_tank
    
    # Specification impacting downstream units (it doesn't matter).
    last_tank.specifications.clear() # Remove specification
    mid_tank.specifications.clear()
    
    @T1.add_specification(run=True, impacted_units=[H3])
    def adjust_flow_rate():
        H3.T = 320
    
    sys.simulate()
    assert H3.outs[0].T == H3.T
    assert T1.specifications[0].path == []
    
    # Two overlapping parallel specifications should make
    # just one parallel specification and one upstream specification
    T1.specifications.clear()
    H1.add_specification(lambda: None, run=True, impacted_units=[T2])
    H2.add_specification(lambda: None, run=True, impacted_units=[T1])
    sys.simulate()
    if sys.unit_path.index(H1) < sys.unit_path.index(T2):
        assert H1.specifications[0].path == []
        assert H2.specifications[0].path == [T1, H1, T2]
        # Net simulation order is ..., T2, H2, T1, H1, T2, H2, ...
    else:
        assert H1.specifications[0].path == [T2, H2, T1]
        assert H2.specifications[0].path == []
    
def test_process_specifications_with_recycles():
    bst.F.set_flowsheet('bifurcated_recycle_loops')
    bst.settings.set_thermo(['Water'], cache=True)
    feed_a = bst.Stream('feed_a', Water=10)
    water_a = bst.Stream('water_a', Water=10)
    recycle_a = bst.Stream('recycle_a')
    P1_a = bst.Pump('P1_a', feed_a)
    P2_a = bst.Pump('P2_a', water_a)
    S1_a = bst.Splitter('S1_a', P2_a-0, split=0.5)
    M1_a = bst.Mixer('M1_a', [P1_a-0, S1_a-1, recycle_a])
    S2_a = bst.Splitter('S2_a', M1_a-0, split=0.5)
    M2_a = bst.Mixer('M2_a', [S1_a-0, S2_a-1])
    S3_a = bst.Splitter('S3_a', M2_a-0, ['', recycle_a], split=0.5)
    feedstock_b = bst.Stream('feedstock_b', Water=1000)
    recycle_b = bst.Stream('recycle_b')
    product_b = bst.Stream('product_b')
    side_product_b = bst.Stream('side_product_b')
    P1_b = bst.Pump('P1_b', feedstock_b)
    P2_b = bst.Pump('P2_b', S2_a-0)
    S1_b = bst.Splitter('S1_b', P2_b-0, split=0.5)
    M1_b = bst.Mixer('M1_b', [P1_b-0, S1_b-1, recycle_b])
    S2_b = bst.Splitter('S2_b', M1_b-0, [side_product_b, ''], split=0.5)
    M2_b = bst.Mixer('M2_b', [S1_b-0, S2_b-1, S3_a-0])
    S3_b = bst.Splitter('S3_b', M2_b-0, [product_b, recycle_b], split=0.5)
    recycle_loop_sys = bst.F.create_system('recycle_loop_sys')
    assert recycle_loop_sys.unit_path == [
        P1_b, P1_a, P2_a, S1_a, M1_a, S2_a, M2_a, 
        S3_a, P2_b, S1_b, M1_b, S2_b, M2_b, S3_b,
    ]
    recycle_loop_sys.simulate()
    x_nested_solution = np.vstack([recycle_a.mol, recycle_b.mol])
    # Test parallel process specification
    P2_b.add_specification(f=lambda: None, run=True, impacted_units=[P1_b])
    recycle_loop_sys.simulate()
    assert recycle_loop_sys.unit_path == [ 
        # Reordered path
        P1_a, P2_a, S1_a, M1_a, S2_a, M2_a, S3_a, 
        P2_b, S1_b, P1_b, M1_b, S2_b, M2_b, S3_b,
    ]
    x_reconfigured_solution = np.vstack([recycle_a.mol, recycle_b.mol])
    assert_allclose(x_nested_solution, x_reconfigured_solution, rtol=2e-2)
    # Test upstream process specification
    M1_b.add_specification(lambda: None, run=True, impacted_units=[S2_a, S3_a])
    recycle_loop_sys.simulate()
    # Make sure path includes upstream subsystem and other units in recycle loop
    assert recycle_loop_sys._to_network() == Network(
        [P1_a,
         P2_a,
         S1_a,
         Network(
            [M1_a,
             S2_a,
             M2_a,
             S3_a],
            recycle=S3_a-1),
         P2_b,
         S1_b,
         P1_b,
         Network(
            [M1_b,
             S2_b,
             M2_b,
             S3_b],
            recycle=S3_b-1)]
    )
    assert M1_b.specifications[0].path == [
        recycle_loop_sys.subsystems[0],
        P2_b,
        S1_b,
        P1_b,
        M2_b,
        S3_b
    ]
    
def test_unit_connections():
    from biorefineries import sugarcane as sc
    sc.load()
    f = sc.flowsheet
    u = f.unit
    all_units = set(sc.sys.units).difference(sc.sys.facilities)
    upstream_units = u.R301.get_upstream_units()
    downstream_units = u.R301.get_downstream_units()
    assert u.R301.neighborhood(1) == {u.T301, u.D301, u.S302, u.H301}
    assert u.R301.neighborhood(2) == {u.S302, u.C301, u.H301, u.M302, u.D301, u.R301,
                                    u.T301, u.M301}
    assert u.R301.neighborhood(200) == u.R301.neighborhood(1000) == all_units
    
    recycle_units = set(sc.sys.find_system(u.R301).units)
    assert recycle_units == upstream_units.intersection(downstream_units)
    ins = tuple(u.R301.ins)
    outs = tuple(u.R301.outs)
    u.R301.disconnect()
    assert not any(u.R301.ins + u.R301.outs)
    ins - u.R301 - outs
    
    class DummyUnit(bst.Unit, isabstract=True, new_graphics=False):
        _N_ins = 2; _N_outs = 2
        
    unit = DummyUnit(None)
    unit.take_place_of(u.R301)
    assert tuple(unit.ins + unit.outs) == (ins + outs)
    assert not any(u.R301.ins + u.R301.outs)
    u.R301.take_place_of(unit)

def test_unit_graphics():
    bst.settings.set_thermo(['Water', 'Ethanol'], cache=True)
    M = bst.Mixer(None, outs=None)
    assert M._graphics.get_inlet_options(M, 2) == {'headport': 'c'}
    assert M._graphics.get_inlet_options(M, 100) == {'headport': 'c'}
    assert M._graphics.get_outlet_options(M, 0) == {'tailport': 'e'}
    
    S = bst.Splitter(None, outs=None, split=0.5)
    
    with pytest.warns(RuntimeWarning):
        assert S._graphics.get_inlet_options(S, 1) == {'headport': 'c'}
    
    with pytest.warns(RuntimeWarning):
        assert M._graphics.get_outlet_options(M, 1) == {'tailport': 'c'}

def test_cost_decorator():
    from biosteam.units.decorators import cost
    bst.settings.set_thermo(['Water'], cache=True)
    @cost('Flow rate', CE=bst.settings.CEPCI, cost=1, n=0.6, lb=2, ub=10, units='kg/hr', BM=2.)
    class A(bst.Unit): pass
    
    feed = bst.Stream('feed', Water=1, units='kg/hr')
    
    # Test when size factor is under lower bound
    A1 = A('A1', ins=feed)
    A1.simulate()
    assert_allclose(A1.purchase_cost, 2 ** 0.6)
    assert_allclose(A1.installed_cost, 2 ** 1.6)
    
    # Test when size factor is between lower and upper bound
    feed.imass['Water'] = 5
    A1.simulate()
    assert_allclose(A1.purchase_cost, 5 ** 0.6)
    assert_allclose(A1.installed_cost, 2 * 5 ** 0.6)
    
    # Test when size factor is above upper bound
    feed.imass['Water'] = 20
    A1.simulate()
    assert_allclose(A1.purchase_cost, 2 * 10 ** 0.6)
    assert_allclose(A1.installed_cost, 4 * 10 ** 0.6)
    
def test_equipment_lifetimes():
    from biorefineries.sugarcane import create_tea
    bst.settings.set_thermo(['Water'], cache=True)
    
    class A(bst.Unit):
        _F_BM_default = {
            'Equipment A': 2,
            'Equipment B': 3,
        }
        _default_equipment_lifetime = {
            'Equipment A': 10,
            'Equipment B': 5,
        }
        def _cost(self):
            purchase_costs = self.baseline_purchase_costs
            purchase_costs['Equipment A'] = 1e6
            purchase_costs['Equipment B'] = 1e5
            
    class B(bst.Unit):
        _F_BM_default = {
            'Equipment A': 2,
        }
        _default_equipment_lifetime = 15
        def _cost(self):
            self.baseline_purchase_costs['Equipment A'] = 1e6

    class C(bst.Unit):
        _F_BM_default = {
            'Equipment A': 4,
        }
        def _cost(self):
            self.baseline_purchase_costs['Equipment A'] = 1e6
            
    @bst.decorators.cost('Flow rate', units='kmol/hr', S=1, BM=3,
                         cost=1e6, n=0.6, CE=bst.CE, lifetime=8)
    class D(bst.Unit): pass
    
    @bst.decorators.cost('Flow rate', units='kmol/hr', S=1, BM=4,
                         cost=1e6, n=0.6, CE=bst.CE, lifetime=20)
    class E(bst.Unit): pass
    
    D_feed = bst.Stream('D_feed', Water=1)
    E_feed = D_feed.copy('E_feed')
    units = [A(None, 'A_feed'), B(None, 'B_feed'), C(None, 'C_feed'), D(None, D_feed), E(None, E_feed)]
    test_sys = bst.System('test_sys', units)
    test_sys.simulate()
    tea = create_tea(test_sys)
    table = tea.get_cashflow_table()
    C_FCI = table['Fixed capital investment [MM$]']
    # Test with lang factor = 3 (default for sugarcane biorefinery)
    cashflows_FCI = [6.12, 9.18, 0.  , 0.  , 0.  , 0.  , 0.  , 0.3 , 0.  , 0.  , 3.  ,
                     0.  , 3.3 , 0.  , 0.  , 0.  , 0.  , 3.3 , 3.  , 0.  , 0.  , 0.  ]
    assert_allclose(C_FCI, cashflows_FCI)
    # Cashflows include maintainance and others, so all entries are not zero
    cashflows = [
       -6120000., -9945000., -4321300., -4321300., -4321300., -4321300.,
       -4321300., -4621300., -4321300., -4321300., -7321300., -4321300.,
       -7621300., -4321300., -4321300., -4321300., -4321300., -7621300.,
       -7321300., -4321300., -4321300., -3556300.
    ]
    assert_allclose(tea.cashflow_array, cashflows)
    
    # Test with bare module costs
    tea.lang_factor = None
    table = tea.get_cashflow_table()
    C_FCI = table['Fixed capital investment [MM$]']
    cashflows_FCI = [
        6.12, 9.18, 0.  , 0.  , 0.  , 0.  , 0.  , 0.3 , 0.  , 0.  , 3.  ,
        0.  , 2.3 , 0.  , 0.  , 0.  , 0.  , 2.3 , 3.  , 0.  , 0.  , 0.  
    ]
    assert_allclose(C_FCI, cashflows_FCI)
    cashflows = [
       -6120000., -9945000., -4321300., -4321300., -4321300., -4321300.,
       -4321300., -4621300., -4321300., -4321300., -7321300., -4321300.,
       -6621300., -4321300., -4321300., -4321300., -4321300., -6621300.,
       -7321300., -4321300., -4321300., -3556300.
    ]
    assert_allclose(tea.cashflow_array, cashflows)

def test_skipping_unit_simulation_with_empty_inlet_streams():
    bst.settings.set_thermo(['Water', 'Ethanol'], cache=True)
    s1 = bst.Stream('s1', Water=10, Ethanol=50)
    s2 = bst.Stream('s2', Water=10)
    M1 = bst.MixTank('M1', ins=(s1, s2))
    
    # Unit must simulate when skip flag is off
    bst.settings.skip_simulation_of_units_with_empty_inlets = False
    M1.simulate()
    assert M1.installed_cost > 0.
    assert M1.utility_cost > 0.
    assert not all([i.isempty() for i in M1.outs])
       
    # Unit must simulate when skip flag is on and inlets are not empty
    bst.settings.skip_simulation_of_units_with_empty_inlets = True
    M1.simulate()
    assert M1.installed_cost > 0.
    assert M1.utility_cost > 0.
    assert not all([i.isempty() for i in M1.outs])
    
    # Unit must have no costs nor outlet flows when skip flag is on and inlets are empty
    bst.settings.skip_simulation_of_units_with_empty_inlets = True
    s1.empty()
    s2.empty()
    M1.simulate()
    assert M1.installed_cost == 0.
    assert M1.utility_cost == 0.
    assert all([i.isempty() for i in M1.outs])
    
    # Unit should not simulate when the skip flag is on 
    # and inlets are empty 
    bst.settings.skip_simulation_of_units_with_empty_inlets = True
    @M1.add_specification
    def f(): raise RuntimeError('unit simulated')
    M1._design = f
    M1._cost = f
    M1.simulate()
    
    # Unit should simulate when the skip flag is on 
    # and inlets are not empty 
    bst.settings.skip_simulation_of_units_with_empty_inlets = False
    with pytest.raises(RuntimeError):
        M1.simulate()
        
if __name__ == '__main__':
    test_auxiliary_unit_owners()
    test_unit_convinience_properties()
    test_unit_inheritance_setup_method()
    test_process_specifications_linear()
    test_process_specifications_with_recycles()
    test_unit_connections()
    test_unit_graphics()
    test_cost_decorator()
    test_equipment_lifetimes()
    test_skipping_unit_simulation_with_empty_inlet_streams()



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
__all__ = ('SystemMesh', 'meshable')

import biosteam as bst
from typing import Sequence

def meshable(obj):
    """Decorator to enable a function or sequence to be accepted to SystemMesh objects."""
    try:
        if not hasattr(obj, 'ins') or not hasattr(obj, 'outs'):
            if isinstance(obj, Sequence):
                obj = bst.MockSystem(obj)
                obj.load_inlet_ports([])
                obj.load_outlet_ports([])
            elif callable(obj):
                if not hasattr(obj, 'ID'): obj.ID = obj.__name__
                if not hasattr(obj, 'ins'): obj.ins = []
                if not hasattr(obj, 'outs'): obj.outs = []
            else:
                assert False
    except:
        raise TypeError(f"'{obj}' is not meshable; only System, MockSystem, "
                        "SystemFactory, Unit, Sequence[Unit], or callable "
                        "objects are meshable")
    return obj


class SystemMesh:
    __slots__ = ('_objects', '_connections', '_inlets', '_outlets')
    
    def __init__(self):
        self._objects = {}
        self._connections = {}
        self._inlets = {}
        self._outlets = {}
    
    def copy(self):
        cls = type(self)
        new = cls.__new__(cls)
        new._objects = self._objects.copy()
        new._connections = self._connections.copy()
        new._inlets = self._inlets.copy()
        new._outlets = self._outlets.copy()
        return new
    
    def add(self, name, obj, number=None, autoconnect=True, **kwargs):
        obj = meshable(obj)
        objs = self._objects
        old_obj = objs[name] if name in objs else None
        objs[name] = (obj, number, kwargs)
        if old_obj:
            self._connections.clear()
            self._inlets.clear()
            self._outlets.clear()
            for i, (j, k, _) in objs.items():
                self._load_obj(i, j, k, autoconnect)
        else:
            self._load_obj(name, obj, number, autoconnect)
        
    def remove(self, name, reconnect=True, **kwargs):
        objs = self._objects
        objs.pop(name)
        self._connections.clear()
        self._inlets.clear()
        self._outlets.clear()
        for i, (j, k, _) in objs.items():
            self._load_obj(i, j, k, reconnect)
        
    def _load_obj(self, name, obj, number, autoconnect):
        isa = isinstance
        isfunc = callable
        for index, inlet in enumerate(obj.ins):
            if isa(inlet, dict):
                ID = inlet.get('ID')
                if ID: self._define_inlet(ID, name, index, autoconnect)
            elif isa(inlet, str):
                ID = inlet
                if ID: self._define_inlet(ID, name, index, autoconnect)
            elif isfunc(inlet):
                ID = inlet.__name__
                self._define_inlet(ID, name, index, autoconnect)
            elif isa(inlet, bst.Stream):
                ID = inlet.ID
                self._define_inlet(ID, name, index, autoconnect)
        for index, outlet in enumerate(obj.outs):
            if isa(outlet, dict):
                ID = outlet.get('ID')
                if ID: self._define_outlet(ID, name, index, autoconnect)
            elif isa(outlet, str):
                ID = outlet
                if ID: self._define_outlet(ID, name, index, autoconnect)
            elif isfunc(outlet):
                ID = outlet.__name__
                self._define_outlet(ID, name, index, autoconnect)
            elif isa(outlet, bst.Stream):
                ID = outlet.ID
                self._define_outlet(ID, name, index, autoconnect)
        
    def _rename(self, dct, name):
        other = dct.pop(name)
        try:
            base, num = name.rsplit('_', 1)
            num = int(num)
        except:
            new_name = name + '_1'
        else:
            new_name = base + '_' + str(num + 1) 
        dct[new_name] = other
        base, num = new_name.rsplit('_', 1)
        name = base + '_' + str(int(num) + 1)
        return name
        
    def _define_inlet(self, name, obj, index, autoconnect=True):
        if obj not in self._objects: raise ValueError('object with name {repr(obj)} does not exist')
        inlets = self._inlets
        value = (str(obj), int(index))
        if name in inlets and value != inlets[name]:
            name = self._rename(inlets, name)
        inlets[name] = value
        if autoconnect and name in self._outlets: self.connect(name, name)
    
    def _define_outlet(self, name, obj, index, autoconnect=True):
        if obj not in self._objects: raise ValueError('object with name {repr(obj)} does not exist')
        outlets = self._outlets
        value = (str(obj), int(index))
        if name in outlets and value != outlets[name]:
            name = self._rename(outlets, name)
        outlets[name] = value
        if autoconnect and name in self._inlets: self.connect(name, name)
        
    def connect(self, outlet, inlet):
        if outlet not in self._outlets:
            raise ValueError('outlet {repr(outlet)} does not exist')
        if inlet not in self._inlets:
            raise ValueError('inlet {repr(inlet)} does not exist')
        self._connections[outlet] = inlet
    
    def disconnect(self, outlet):
        if outlet not in self._outlets:
            raise ValueError('outlet {repr(outlet)} does not exist')
        self._connections.pop(outlet)
    
    def __call__(self, ID, **streams):
        connections = self._connections
        inlets = self._inlets 
        outlets = self._outlets
        objs = {}
        with bst.System(ID) as sys:
            old_units = []
            for name, (obj, N, kwargs) in self._objects.items(): 
                if isinstance(obj, bst.SystemFactory):
                    objs[name] = obj(name, area=N, mockup=True, **kwargs)
                elif isinstance(obj, (bst.System, bst.MockSystem)):
                    objs[name] = system = obj
                    if N is not None: # Rename if given area number
                        for i in system.units: i.ID = N
                    old_units.extend(system.units)
                    # Add units to context management
                    bst.main_flowsheet.unit.context_levels[-1].extend(system.units)
                elif isinstance(obj, bst.Unit):
                    objs[name] = unit = obj
                    if N is not None: unit.ID = N # Rename if given area number
                    old_units.append(unit)
                    # Add units to context management
                    bst.main_flowsheet.unit.context_levels[-1].append(unit)
                elif callable(obj):
                    obj(**kwargs)
                else:
                    raise RuntimeError(f"invalid object type '{type(obj).__name__}'")
            for name, stream in streams.items():
                if name in inlets:
                    objname, index = inlets[name]
                    objs[objname].ins[index] = stream
                elif name in outlets: 
                    objname, index = outlets[name]
                    objs[objname].outs[index] = stream
                else: raise ValueError(f'no inlet or outlet named {repr(name)}')
            for outname, inname in connections.items():
                upstream, index = outlets[outname]
                stream = objs[upstream].outs[index]
                downstream, index = inlets[inname]
                objs[downstream].ins[index] = stream
            for i in old_units:
                if getattr(i, 'autopopulate', False): i.ins.clear()
        return sys
    
    def show(self):
        objs, connections, inlets, outlets = (self._objects, self._connections, self._inlets, self._outlets)
        info = f"{type(self).__name__}:"
        def get_description(obj):
            if isinstance(obj, bst.SystemFactory):
                return obj.f.__name__
            elif hasattr(obj, 'ID') and isinstance(obj.ID, str):
                return obj.ID
            elif isinstance(obj, bst.MockSystem):
                if len(obj.units) < 4:
                    units = obj.units
                    return ', '.join([i.ID for i in units])
                else:
                    units = obj.units[:4]
                    return ', '.join([i.ID for i in units]) + ', ...'
        fields = {i: {'description': get_description(j), 'number': N, 'ins': {}, 'outs': {}}
                  for i, (j, N, _) in objs.items()}
        connections = connections.copy()
        for i, j in connections.items(): connections[j] = i
        for name, (obj, index) in inlets.items():
            dct = fields[obj]['ins']
            if name in connections:
                outname = connections[name]
                outobj, outindex = outlets[outname]
                description = f"{name} from {outobj}-{outindex}"
            else:
                stream = objs[obj][0].ins[index]
                if hasattr(stream, 'source') and stream.source:
                    description = f"{name} from {stream.source}-{stream.source.outs.index(stream)}"
                else:
                    description = name
            dct[index] = description
        for name, (obj, index) in outlets.items():
            dct = fields[obj]['outs']
            if name in connections:
                inname = connections[name]
                inobj, inindex = inlets[inname]
                description = f"{name} to {inindex}-{inobj}"
            else:
                stream = objs[obj][0].outs[index]
                if hasattr(stream, 'sink') and stream.sink:
                    description = f"{name} to {stream.sink}-{stream.sink.ins.index(stream)}"
                else:
                    description = name
            dct[index] = description
        keys = ('ins', 'outs')
        section_size = max([len(i) for i in keys]) + 1 # for spaces
        for name, dct in fields.items():
            if dct['number']:
                if dct['description']:
                    info += f"\n{name} ({dct['number']}; {dct['description']})"
                else:
                    info += f"\n{name} ({dct['number']})"
            elif dct['description']:
                info += f"\n{name} ({dct['description']})"
            else:
                info += f"\n{name}"
            for key in keys:
                descriptions = dct[key]
                if descriptions:
                    descriptions = sorted([f"[{i}] {j}" for i, j in descriptions.items()])
                    section = key + (section_size - len(key)) * ' '
                    info += '\n' + section
                    spaces = section_size * ' '
                    info += ('\n' + spaces).join(descriptions)
        print(info)
        
    _ipython_display_ = show
        
        


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
#               2023-,     Yalin Li <mailto.yalin.li@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import biosteam as bst

__all__ = (
    'create_facilities', 
    'create_all_facilities',
    'create_coheat_and_power_system',
)

def create_all_facilities(
    feedstock=None,
    CIP_over_feedstock=None,
    plant_air_over_feedstock=None,
    fire_water_over_feedstock=None,
    recycle_process_water_streams=None,
    treated_water_streams=None,
    CT=True,
    CWP=True,
    CIP=True,
    FWT=True,
    ADP=True,
    WWT=True,
    CHP=True,
    HXN=True,
    PWC=True,
    CT_kwargs=None,
    CWP_kwargs=None,
    CIP_kwargs=None,
    FWT_kwargs=None,
    ADP_kwargs=None,
    WWT_kwargs=None,
    CHP_kwargs=None,
    HXN_kwargs=None,
    PWC_kwargs=None,
    area=None,
    blowdown_recycle=False,
):
    """Create all facilities by default."""
    return create_facilities(
        feedstock, CIP_over_feedstock, plant_air_over_feedstock, fire_water_over_feedstock,
        recycle_process_water_streams, treated_water_streams,
        CT=CT,
        CWP=CWP,
        CIP=CIP,
        FWT=FWT,
        ADP=ADP,
        WWT=WWT,
        CHP=CHP,
        HXN=HXN,
        PWC=PWC,
        CT_kwargs=CT_kwargs,
        CWP_kwargs=CWP_kwargs,
        CIP_kwargs=CIP_kwargs,
        FWT_kwargs=FWT_kwargs,
        ADP_kwargs=ADP_kwargs,
        WWT_kwargs=WWT_kwargs,
        CHP_kwargs=CHP_kwargs,
        HXN_kwargs=HXN_kwargs,
        PWC_kwargs=PWC_kwargs,
        area=area,
        blowdown_recycle=blowdown_recycle,
    )

def create_facilities(
        feedstock=None,
        CIP_over_feedstock=None,
        plant_air_over_feedstock=None,
        fire_water_over_feedstock=None,
        recycle_process_water_streams=None,
        treated_water_streams=None,
        CT=None,
        CWP=None,
        CIP=None,
        FWT=None,
        ADP=None,
        WWT=None,
        CHP=None,
        HXN=None,
        PWC=None,
        CT_kwargs=None,
        CWP_kwargs=None,
        CIP_kwargs=None,
        FWT_kwargs=None,
        ADP_kwargs=None,
        WWT_kwargs=None,
        CHP_kwargs=None,
        HXN_kwargs=None,
        PWC_kwargs=None,
        area=None,
        blowdown_recycle=False,
    ):
    """Create facilities specified."""
    with bst.MockSystem() as sys:
        kwargs = (CHP_kwargs, WWT_kwargs, CT_kwargs, CWP_kwargs, CIP_kwargs,
                  FWT_kwargs, ADP_kwargs, HXN_kwargs, PWC_kwargs)
        (CHP_kwargs, WWT_kwargs, CT_kwargs, CWP_kwargs, CIP_kwargs,
         FWT_kwargs, ADP_kwargs, HXN_kwargs, PWC_kwargs) = kwargs = [
             {} if i is None else i for i in kwargs
        ]
        if area:
            for i in kwargs[:2]: 
                if 'area' not in i: i['area'] = area
            for i in kwargs[2:]: 
                if 'ID' not in i: i['ID'] = area
        if CT: bst.facilities.CoolingTower(**CT_kwargs)
        if CWP: bst.facilities.ChilledWaterPackage(**CWP_kwargs)
        # Wastewater treatment system, two configurations with the conventional one as the default
        if WWT:
            if 'autopopulate' not in WWT_kwargs:
                WWT_kwargs = WWT_kwargs.copy()
                WWT_kwargs['autopopulate'] = True
            bst.create_wastewater_treatment_system(**WWT_kwargs, mockup=True)
        if HXN: bst.HeatExchangerNetwork(**HXN_kwargs)
        if feedstock is not None:
            if CIP:
                CIP = bst.Stream('CIP', Water=126, units='kg/hr')
                CIP_package = bst.CIPpackage(ins=CIP, **CIP_kwargs)
                CIP_package.CIP_over_feedstock = 0.00121 if CIP_over_feedstock is None else CIP_over_feedstock
                @CIP_package.add_specification(run=True)
                def adjust_CIP(): CIP.imass['Water'] = feedstock.F_mass * CIP_package.CIP_over_feedstock
            if ADP:
                plant_air = bst.Stream('plant_air', N2=83333, units='kg/hr')
                ADP = bst.AirDistributionPackage(ins=plant_air, **ADP_kwargs)
                ADP.plant_air_over_feedstock = 0.8 if plant_air_over_feedstock is None else plant_air_over_feedstock
                @ADP.add_specification(run=True)
                def adjust_plant_air(): plant_air.imass['N2'] = feedstock.F_mass * ADP.plant_air_over_feedstock
            if FWT:
                fire_water = bst.Stream('fire_water', Water=8343, units='kg/hr')
                FT = bst.FireWaterTank(ins=fire_water, **FWT_kwargs)
                FT.fire_water_over_feedstock = 0.08 if fire_water_over_feedstock is None else fire_water_over_feedstock
                @FT.add_specification(run=True)
                def adjust_fire_water(): fire_water.imass['Water'] = feedstock.F_mass * FT.fire_water_over_feedstock
        if CHP:
            if 'autopopulate' not in CHP_kwargs:
                CHP_kwargs = CHP_kwargs.copy()
                CHP_kwargs['autopopulate'] = True
            create_coheat_and_power_system(mockup=True, **CHP_kwargs)
        if blowdown_recycle:
            units = bst.main_flowsheet.unit.get_context_level(0)
            blowdown_to_wastewater = bst.Stream('blowdown_to_wastewater')
            bst.BlowdownMixer(
                area or '',
                [i.blowdown_water for i in units if hasattr(i, 'blowdown_water')],
                blowdown_to_wastewater
            )
        if PWC:
            process_water_mixer = bst.Mixer(area or '',
                ins=[] if recycle_process_water_streams is None else recycle_process_water_streams
            )
            process_water_mixer.autopopulate = True if recycle_process_water_streams is None else False 
            
            @process_water_mixer.add_specification(run=True)
            def autopopulate_recycle_process_water_streams():
                if process_water_mixer.autopopulate and not process_water_mixer.ins:
                    streams = bst.get_streams_from_context_level(0)
                    process_water_mixer.ins.extend([i for i in streams if i.isproduct() and 'process_water' in i.ID])
            
            process_water = process_water_mixer.outs[0]
            if treated_water_streams is None:
                units = bst.main_flowsheet.unit.get_context_level(0)
                treated_water_streams = [i.RO_treated_water for i in units if hasattr(i, 'RO_treated_water')]
            treated_water_mixer = bst.Mixer(area or '', ins=treated_water_streams)
            treated_water = treated_water_mixer.outs[0]
            bst.ProcessWaterCenter(ins=[treated_water, 'makeup_RO_water', process_water, 'makeup_process_water'], **PWC_kwargs)
    return sys.units

@bst.SystemFactory(
    ID='CHP_sys',
    ins=['makeup_water', 'natural_gas', 'FGD_lime', 'boiler_chemicals', 'air'],
    outs=['emissions', 'blowdown', 'ash'],
    fixed_ins_size=False,
)
def create_coheat_and_power_system(
        ins, outs, combustible_slurries=None, combustible_gases=None, 
        autopopulate=None, slurry_feed=True, gas_feed=True, **kwargs
    ):
    makeup_water, natural_gas, lime, boiler_chemicals, air = ins
    if 'fuel_source' in kwargs: natural_gas.ID = kwargs['fuel_source'].lower()
    if slurry_feed and autopopulate or combustible_slurries:
        slurry = bst.Stream('to_boiler')
        slurry_mixer = bst.Mixer(ins=combustible_slurries or [], outs=slurry)
        slurry_mixer.register_alias('slurry_mixer')
        slurry = slurry_mixer-0
    else:
        slurry = None
    if gas_feed and autopopulate or combustible_gases:
        gas_to_boiler = bst.Stream('gas_to_boiler')
        gas_mixer = bst.Mixer(ins=combustible_gases or [], outs=gas_to_boiler)
        gas_mixer.register_alias('gas_mixer')
        gas = gas_mixer-0
        gas.phase = 'g'
    else:
        gas = None
    BT = bst.BoilerTurbogenerator(
        ins=[slurry, gas, 
             makeup_water, natural_gas, 
             lime, boiler_chemicals, air],
        outs=outs,
        **kwargs,
    )
    BT.autopopulate = False if autopopulate is None else autopopulate
    
    @BT.add_specification(run=True)
    def autopopulate_combustibles():
        if BT.autopopulate and not (slurry_feed and slurry_mixer.ins) and not (gas_feed and gas_mixer.ins):
            streams = bst.FreeProductStreams(BT.system.streams)
            if slurry_feed: 
                slurry_mixer.ins.extend([i for i in streams.combustible_slurries if i is not slurry])
                slurry_mixer.run_until(BT)
            if gas_feed: 
                gas_mixer.ins.extend([i for i in streams.combustible_gases if i is not gas])
                gas_mixer.run_until(BT)
        else:
            BT.autopopulate = False



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
General functional algorithms for the design and purchase cost estimation
of agitators.

References
----------
.. [1] Seider, W. D., Lewin,  D. R., Seader, J. D., Widagdo, S., Gani, R.,
    & Ng, M. K. (2017). Product and Process Design Principles. Wiley.
    Cost Accounting and Capital Cost Estimation (Chapter 16)

"""
from numba import njit
import biosteam as bst
__all__ = (
    'compute_closed_vessel_turbine_purchase_cost',
)

@njit(cache=True)
def _compute_closed_vessel_turbine_purchase_cost(power, CE):
    C_p = 4105 * power ** 0.57
    return CE / 567 * C_p

def compute_closed_vessel_turbine_purchase_cost(power):
    """
    Return the purchase cost [Cp; in USD] of a turbine for closed vessels.
    
    Parameters
    ----------
    power : float
        Power [hp].
    
    Examples
    --------
    >>> compute_closed_vessel_turbine_purchase_cost(power=10)
    15264.97
    
    Notes
    -----
    The purchase cost is given by [1]_. See source code for details.
    The purchase cost is scaled according to BioSTEAM's Chemical
    Plant Cost Index, `biosteam.CE`.
    
    """
    return _compute_closed_vessel_turbine_purchase_cost(power, bst.CE)


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""

import pytest
import biosteam as bst

def test_rotary_vane_vacuum_pump():

    bst.settings.set_thermo(['Water', 'CH4'])
    gas = bst.Stream('gas', CH4=2, Water=0.1)
    vac = bst.VacuumSystem(F_mass=gas.F_mass, F_vol=gas.F_vol, 
        P_suction=31325., vessel_volume=1.9e-3)
    assert "Rotary-vane pump, one stage" in vac.baseline_purchase_costs
    #!!! TODO: add test to purchase cost after updating CEPCI
    
    vac = bst.VacuumSystem(F_mass=gas.F_mass*10, F_vol=gas.F_vol*10, 
        P_suction=31325., vessel_volume=1.9e-3)
    assert "Rotary-vane pump, two stage" in vac.baseline_purchase_costs
    pass

if __name__ == '__main__':
    test_rotary_vane_vacuum_pump()





# -*- coding: utf-8 -*-
# HXN: The automated Heat Exchanger Network design package.
# Copyright (C) 2020-, Sarang Bhagwat <sarangb2@illinois.edu>
# 
# This module is under the UIUC open-source license. See 
# github.com/sarangbhagwat/hxn/blob/master/LICENSE.txt
# for license details.
"""
Created on Sat May  2 16:44:24 2020

@author: sarangbhagwat
"""
import numpy as np
import biosteam as bst
from warnings import warn

__all__ = ('StreamLifeCycle', 'synthesize_network')

class LifeStage:
        
    def __init__(self, unit, index):
        self.unit = unit
        self.index = index
    
    @property
    def s_in(self): return self.unit.ins[self.index]
    
    @property
    def s_out(self): return self.unit.outs[self.index]
    
    @property
    def H_in(self): return self.s_in.H
    
    @property
    def H_out(self): return self.s_out.H
    
    def _info(self, N_tabs=1):
        tabs = N_tabs*'\t'
        return (f"{type(self).__name__}: {self.unit.ID}\n"
                + tabs + f"H_in = {self.H_in:.3g} kJ\n"
                + tabs + f"H_out = {self.H_out:.3g} kJ")
                
    def __repr__(self):
        return (f"<{type(self).__name__}: {repr(self.unit)}, H_in = {round(self.H_in, 4):.3g} kJ, H_out = {round(self.H_out, 4):.3g} kJ>")
        
    def show(self):
        print(self._info())
    _ipython_display_ = show
        
        
class StreamLifeCycle:
    
    def __init__(self, index, cold):
        self.index = index
        self.name = 's_%s'%index
        self.cold = cold
        self.life_cycle = None
        
    def get_relevant_units(self, index, new_HXs, new_HX_utils):
        new_HXs_relevant = [hx for hx in new_HXs if '_%s_'%index in hx.ID]
        new_HX_utils_relevant = [hx for hx in new_HX_utils if '_%s_'%index in hx.ID]
        return new_HXs_relevant, new_HX_utils_relevant
        
    def get_life_cycle(self, new_HXs, new_HX_utils):
        index = self.index
        name = self.name
        cold = self.cold
        new_HXs_relevant, new_HX_utils_relevant =\
            self.get_relevant_units(index, new_HXs, new_HX_utils)
        life_cycle = (
            [LifeStage(unit, 0) for unit in new_HXs_relevant if name + '_' in unit.ins[0].ID]
            + [LifeStage(unit, 1) for unit in new_HXs_relevant if name + '_' in unit.ins[1].ID]
            + [LifeStage(unit, 0) for unit in new_HX_utils_relevant if name + '_' in unit.ins[0].ID]
        )
        life_cycle.sort(key = lambda pt: pt.H_in, reverse = not cold)
        self.life_cycle = life_cycle
        return life_cycle
        
    def __repr__(self):
        life_cycle = self.life_cycle
        cold = self.cold
        if not self.life_cycle:
            return 'Not initialized; run StreamLifeCycle.get_life_cycle or\
                  HX_Network.get_stream_life_cycles first.' 
        else:
            index = self.index
            name = 'Stream_%s'%index
            strtype = 'cold' if cold else 'hot'
            rep = ''
            for LifeStage in life_cycle:
                line = '\t\t' + repr(LifeStage) + '\n'
                rep += line
            rep = '<StreamLifeCycle: ' + name + ', ' + strtype  + '\n\tlife_cycle = [\n' +  rep[:-1] + '\n\t]>'
            return rep
        
    def show(self):
        info = repr(self).replace('[', '').replace(']', '').replace('life_cycle =', 'life_cycle:')
        print(info[1:-1])
        
    _ipython_display_ = show


class Working_Life_Cycle:
    
    def __init__(self, index, cold):
        self.index = index
        self.name = 's_%s'%index
        self.cold = cold
        self.life_cycle = life_cycle = {}
        life_cycle['cold_side'] = []
        life_cycle['hot_side'] = []
        
    def add_stage(self, s_in, s_out, side):
        self.life_cycle[side].append(LifeStage(s_in, s_out))
    
    def sort_stages(self):
        life_cycle = self.life_cycle
        reverse = not self.cold
        life_cycle['cold_side'].sort(key = lambda stage: stage.H, reverse = reverse)
        life_cycle['cold_side'].sort(key = lambda stage: stage.H, reverse = reverse)
    
    def get_sorted_life_cycle(self):
        self.sort_stages()
        return self.life_cycle
    

def temperature_interval_pinch_analysis(hus, 
                                        T_min_app=10, 
                                        force_ideal_thermo=False,
                                        sort_hus_by_T=False):
    hx_utils = hus
    hus_heating = [hu for hu in hx_utils if hu.duty > 0]
    hus_cooling = [hu for hu in hx_utils if hu.duty < 0]
    if sort_hus_by_T:
        hus_heating.sort(key=lambda i: i.unit.ins[0].T, reverse=True)
        hus_cooling.sort(key=lambda i: i.unit.ins[0].T)
    hx_utils_rearranged = hus_heating + hus_cooling
    hxs = [hu.unit for hu in hx_utils_rearranged]
    if force_ideal_thermo:
        streams_inlet = [hx.ins[0] for hx in hxs]
        streams_quenched = [i.outs[0] for i in hxs]
        streams_inlet = [i.copy(thermo=i.thermo.ideal()) for i in streams_inlet]
        streams_quenched = [i.copy(thermo=i.thermo.ideal()) for i in streams_quenched]
    else:
        streams_inlet = [hx.ins[0].copy() for hx in hxs]
        streams_quenched = [i.outs[0].copy() for i in hxs]
    for i in streams_quenched: i.vle(H=i.H, P=i.P)
    for i in range(len(streams_inlet)):
        stream = streams_inlet[i]
        ID = 'Util_%s'%i
        stream.ID = 's_%s__%s'%(i,ID)
    N_heating = len(hus_heating)
    is_cold_stream_index = lambda x: x < N_heating
    T_in_arr = np.array([stream.T for stream in streams_inlet])
    T_out_arr = np.array([i.T for i in streams_quenched])
    adj_T_in_arr = T_in_arr.copy()
    # adj_T_in_arr[:N_heating] -= T_min_app
    adj_T_in_arr[N_heating:] -= T_min_app
    adj_T_out_arr = T_out_arr.copy()
    # adj_T_out_arr[:N_heating] -= T_min_app
    adj_T_out_arr[N_heating:] -= T_min_app
    T_changes_tuples = list(zip(adj_T_in_arr, adj_T_out_arr))
    all_Ts_descending = [*adj_T_in_arr, *adj_T_out_arr]
    all_Ts_descending.sort(reverse=True)
    stream_indices_for_T_intervals =\
        {(all_Ts_descending[i], all_Ts_descending[i+1]):[]\
            for i in range(len(all_Ts_descending)-1)}
    H_for_T_intervals = dict.fromkeys(stream_indices_for_T_intervals, 0)
    cold_indices = list(range(N_heating))
    hot_indices = list(range(N_heating, len(hxs)))
    indices = cold_indices + hot_indices
    for i in range(len(all_Ts_descending)-1):
        T_start = all_Ts_descending[i]
        T_end = all_Ts_descending[i+1]
        for stream_index in indices:
            T1, T2 = T_changes_tuples[stream_index]
            if (T1 >= T_start and T2 <= T_end) or (T2 >= T_start and T1 <= T_end):
                multiplier = -1 if is_cold_stream_index(stream_index) else 1
                stream = streams_inlet[stream_index].copy()
                if stream.T != T_start: stream.vle(T = T_start, P = stream.P)
                H1 = stream.H
                try:
                    stream.vle(T = T_end, P = stream.P)
                except:
                    warn(f"could not solve VLE for {repr(stream)} at {repr(hxs[stream_index].owner)}", RuntimeWarning)
                H2 = stream.H
                H = multiplier*(H1 - H2)
                H_for_T_intervals[(T_start, T_end)] += H
        
    res_H_vector = []
    prev_res_H = 0
    for interval, H in H_for_T_intervals.items():
        res_H_vector.append(prev_res_H + H)
        prev_res_H = res_H_vector[len(res_H_vector)-1]
    hot_util_load = - min(res_H_vector)
    # assert hot_util_load>= 0, 'Hot utility load is negative'
    if not hot_util_load>=0:
        warn(f"Hot utility load is negative: {hot_util_load}", RuntimeWarning)
    # print(hot_util_load)
    # the lower temperature of the temperature interval for which the res_H is minimum
    pinch_cold_stream_T = all_Ts_descending[res_H_vector.index(-hot_util_load)+1]
    pinch_hot_stream_T = pinch_cold_stream_T + T_min_app
    cold_util_load = res_H_vector[len(res_H_vector)-1] + hot_util_load
    # assert cold_util_load>=0, 'Cold utility load is negative'
    if not cold_util_load>=0:
        warn(f"Cold utility load is positive: {cold_util_load}", RuntimeWarning)
    pinch_T_arr = []
    for i in cold_indices:
        if T_in_arr[i] > pinch_cold_stream_T:
            pinch_T_arr.append(T_in_arr[i])
        elif T_out_arr[i] < pinch_cold_stream_T:
            pinch_T_arr.append(T_out_arr[i])
        else:
            pinch_T_arr.append(pinch_cold_stream_T)
    for i in hot_indices:
        if T_in_arr[i] < pinch_hot_stream_T:
            pinch_T_arr.append(T_in_arr[i])
        elif T_out_arr[i] > pinch_hot_stream_T:
            pinch_T_arr.append(T_out_arr[i])
        else:
            pinch_T_arr.append(pinch_hot_stream_T)
    pinch_T_arr = np.array(pinch_T_arr)
    # print(pinch_T_arr, hot_util_load, cold_util_load,)
    return pinch_T_arr, hot_util_load, cold_util_load, T_in_arr, T_out_arr,\
           hxs, hot_indices, cold_indices, indices, streams_inlet, hx_utils_rearranged, \
           streams_quenched
            
        
def load_duties(streams, streams_quenched, pinch_T_arr, T_out_arr, indices, is_cold, Q_hot_side, Q_cold_side):
    for index in indices:
        stream = streams[index].copy()
        H_in = stream.H
        stream.vle(T = pinch_T_arr[index], P = stream.P)
        H_pinch = stream.H
        H_out = streams_quenched[index].H
        if not is_cold(index):
            dH1 = abs(H_pinch - H_in)
            dH2 = abs(H_out - H_pinch)
            if abs(dH1)<0.01: dH1 = 0
            if abs(dH2)<0.01: dH2 = 0
            Q_hot_side[index] = ['cool', dH1]
            Q_cold_side[index] = ['cool', dH2]
        else:
            dH1 = H_out - H_pinch
            dH2 = H_pinch - H_in
            if abs(dH1)<0.01: dH1 = 0
            if abs(dH2)<0.01: dH2 = 0
            Q_hot_side[index] = ['heat', dH1]
            Q_cold_side[index] = ['heat', dH2]
            
            
def get_T_transient(pinch_T_arr, indices, T_in_arr):
    T_transient = pinch_T_arr.copy()
    T_transient[indices] = T_in_arr[indices]
    return T_transient

def synthesize_network(hus, T_min_app=5., Qmin=1e-3, force_ideal_thermo=False,
                       avoid_recycle=False, sort_hus_by_T=False):  
    pinch_T_arr, hot_util_load, cold_util_load, T_in_arr, T_out_arr,\
        hxs, hot_indices, cold_indices, indices, streams_inlet, hx_utils_rearranged, \
        streams_quenched = temperature_interval_pinch_analysis(hus, T_min_app, force_ideal_thermo,
                                                               sort_hus_by_T)        
    H_out_arr = [i.H for i in streams_quenched]
    duties = np.array([abs(hx.Q)  for hx in hxs])
    dTs = np.abs(T_in_arr - T_out_arr)
    dTs[dTs == 0.] = 1e-12
    C_flow_vector = duties/dTs
    Q_hot_side = {}
    Q_cold_side = {}
    stream_HXs_dict = {i:[] for i in indices}
    is_cold = lambda x: x in cold_indices
    load_duties(streams_inlet, streams_quenched, pinch_T_arr, T_out_arr, indices, is_cold, Q_hot_side, Q_cold_side)
    matches_hs = {i: [] for i in cold_indices}
    matches_cs = {i: [] for i in hot_indices}
    HXs_hot_side = []
    HXs_cold_side = []
    streams_transient_cold_side = streams_inlet
    streams_transient_hot_side = [i.copy() for i in streams_inlet]
    for i in hot_indices:
        s = streams_transient_cold_side[i]
        if s.T != pinch_T_arr[i]:
            s.vle(T=pinch_T_arr[i], P=s.P)
    for i in cold_indices:
        s = streams_transient_hot_side[i]
        if s.T != pinch_T_arr[i]:
            s.vle(T=pinch_T_arr[i], P=s.P)
    
    def get_stream_at_H_max(cold):
        s_cs = streams_transient_cold_side[cold]
        s_hs = streams_transient_hot_side[cold]
        return s_cs if s_cs.H > s_hs.H else s_hs
    
    def get_stream_at_H_min(hot):
        s_cs = streams_transient_cold_side[hot]
        s_hs = streams_transient_hot_side[hot]
        return s_cs if s_cs.H < s_hs.H else s_hs
    
    def get_T_transient_cold_side(index):
        return streams_transient_cold_side[index].T
    
    def get_T_transient_hot_side(index):
        return streams_transient_hot_side[index].T
    
    attempts = set()
    success = set()
    # ------------- Cold side design ------------- # 
    unavailables = set([i for i in hot_indices if T_out_arr[i] >= pinch_T_arr[i]])
    unavailables.update([i for i in cold_indices if T_in_arr[i] >= pinch_T_arr[i]])
    for hot in hot_indices:
        stream_quenched = False
        potential_matches = []
        for cold in cold_indices:
            if (C_flow_vector[hot]>= C_flow_vector[cold] and
                    get_T_transient_cold_side(hot) > get_T_transient_cold_side(cold) + T_min_app and
                    (hot not in unavailables) and (cold not in unavailables) and
                    (cold not in matches_cs[hot]) and (cold in cold_indices)): 
                potential_matches.append(cold)
        potential_matches = sorted(
            potential_matches, 
            key = lambda pot_cold: min(C_flow_vector[hot], C_flow_vector[pot_cold])
                                    * (get_T_transient_cold_side(hot) 
                                       - get_T_transient_cold_side(pot_cold)
                                       - T_min_app),
            reverse = True
        )
        for cold in potential_matches:
            match = (hot, cold)
            ID = 'HX_%s_%s_cs' % match
            if ID in attempts or (avoid_recycle and match in success): continue
            attempts.add(ID)
            hot_stream = streams_transient_cold_side[hot].copy()
            cold_stream = streams_transient_cold_side[cold].copy()
            
            hot_stream.ID = 's_%s__%s'%(hot,ID)
            cold_stream.ID = 's_%s__%s'%(cold,ID)
            hot_out = hot_stream.copy('%s__s_%s'%(ID,hot))
            cold_out = cold_stream.copy('%s__s_%s'%(ID,cold))
            H_lim = H_out_arr[hot]
            new_HX = bst.units.HXprocess(ID = ID, ins = (hot_stream, cold_stream),
                     outs = (hot_out, cold_out), H_lim0 = H_lim, 
                     T_lim1 = pinch_T_arr[cold], dT = T_min_app,
                     thermo = hot_stream.thermo)
            try: new_HX._run()
            except: continue
            if abs(new_HX.Q )< Qmin: continue
            success.add(match)
            HXs_cold_side.append(new_HX)
            stream_HXs_dict[hot].append(new_HX)
            stream_HXs_dict[cold].append(new_HX)
            Q_cold_side[hot][1] -= new_HX.Q
            Q_cold_side[cold][1] -= new_HX.Q
            streams_transient_cold_side[hot] = new_HX.outs[0]
            streams_transient_cold_side[cold] = new_HX.outs[1]
            H_out = new_HX.outs[0].H
            assert H_out - new_HX.ins[0].H <= 0.
            stream_quenched = H_out < H_lim or np.allclose(H_out, H_lim)
            matches_cs[hot].append(cold)
            if stream_quenched:
                break
    
    # ------------- Hot side design ------------- #                            
    unavailables = set([i for i in hot_indices if T_in_arr[i] <= pinch_T_arr[i]])
    unavailables.update([i for i in cold_indices if T_out_arr[i] <= pinch_T_arr[i]])
    
    for cold in cold_indices:
        potential_matches = []
        for hot in hot_indices:
            if ((cold in matches_cs and hot in matches_cs[cold])
                or (cold in matches_hs and hot in matches_hs[cold])):
                break
            if (C_flow_vector[cold]>= C_flow_vector[hot] and
                    get_T_transient_hot_side(hot) > get_T_transient_hot_side(cold) + T_min_app and
                    (hot not in unavailables) and (cold not in unavailables) and
                    (hot not in matches_hs[cold]) and (hot in hot_indices)):
                potential_matches.append(hot)
                
        potential_matches = sorted(potential_matches, key = lambda x:
                                    (min(C_flow_vector[cold], C_flow_vector[x])
                                        * ( get_T_transient_hot_side(x)
                                          - get_T_transient_hot_side(cold) - T_min_app)),
                                    reverse = True)
        stream_quenched = False
        for hot in potential_matches:
            match = (hot, cold)
            ID = 'HX_%s_%s_hs' % (cold, hot)
            if ID in attempts or (avoid_recycle and match in success): continue
            attempts.add(ID)
            hot_stream = streams_transient_hot_side[hot].copy()
            cold_stream = streams_transient_hot_side[cold].copy()
            cold_stream.ID = 's_%s__%s'%(cold,ID)
            hot_stream.ID = 's_%s__%s'%(hot,ID)
            hot_out = hot_stream.copy('%s__s_%s'%(ID,hot))
            cold_out = cold_stream.copy('%s__s_%s'%(ID,cold))
            H_lim = H_out_arr[cold]
            new_HX = bst.units.HXprocess(ID = ID, ins = (cold_stream, hot_stream),
                     outs = (cold_out, hot_out), H_lim0 = H_lim,
                     T_lim1 = pinch_T_arr[hot], dT = T_min_app,
                     thermo = hot_stream.thermo)
            try: new_HX._run()
            except: continue
            if abs(new_HX.Q)< Qmin: continue
            success.add(match)
            HXs_hot_side.append(new_HX)
            stream_HXs_dict[hot].append(new_HX)
            stream_HXs_dict[cold].append(new_HX)
            Q_hot_side[hot][1] -= new_HX.Q
            Q_hot_side[cold][1] -= new_HX.Q
            streams_transient_hot_side[cold] = new_HX.outs[0]
            streams_transient_hot_side[hot] = new_HX.outs[1]
            H_out = new_HX.outs[0].H
            assert H_out - new_HX.ins[0].H >= 0.
            stream_quenched = H_out > H_lim or np.allclose(H_out, H_lim)
            matches_hs[cold].append(hot)
            if stream_quenched:
                break
    
    # Offset heating requirement on cold side
    for cold in cold_indices:
        if Q_cold_side[cold][0]=='heat' and Q_cold_side[cold][1]>0:
            for hot in hot_indices:
                match = (hot, cold)
                ID = 'HX_%s_%s_cs' % match
                if ID in attempts or (avoid_recycle and match in success): continue
                attempts.add(ID)
                T_cold_in = get_T_transient_cold_side(cold)
                T_hot_in = get_T_transient_cold_side(hot)
                if (Q_cold_side[hot][0]=='cool' and Q_cold_side[hot][1]>0 and
                        T_hot_in - T_cold_in >= T_min_app):
                    hot_stream = streams_transient_cold_side[hot].copy()
                    cold_stream = streams_transient_cold_side[cold].copy()
                    hot_stream.ID = 's_%s__%s'%(hot,ID)
                    cold_stream.ID = 's_%s__%s'%(cold,ID)
                    hot_out = hot_stream.copy('%s__s_%s'%(ID,hot))
                    cold_out = cold_stream.copy('%s__s_%s'%(ID,cold))
                    new_HX = bst.units.HXprocess(ID = ID, ins = (hot_stream, cold_stream),
                             outs = (hot_out, cold_out), H_lim0 = H_out_arr[hot],
                             T_lim1 = T_out_arr[cold], dT = T_min_app,
                             thermo = hot_stream.thermo)
                    try: new_HX._run()
                    except: continue
                    if abs(new_HX.Q )< Qmin: continue
                    success.add(match)
                    HXs_cold_side.append(new_HX)
                    stream_HXs_dict[hot].append(new_HX)
                    stream_HXs_dict[cold].append(new_HX)                    
                    Q_cold_side[hot][1] -= new_HX.Q
                    Q_cold_side[cold][1] -= new_HX.Q                 
                    streams_transient_cold_side[hot] = new_HX.outs[0]
                    streams_transient_cold_side[cold] = new_HX.outs[1]
                    matches_cs[hot].append(cold)
                 
    # Offset cooling requirement on hot side  
    for hot in hot_indices:
        stream_quenched = False
        if Q_hot_side[hot][0]=='cool' and Q_hot_side[hot][1]>0:
            for cold in cold_indices:
                match = (hot, cold)
                ID = 'HX_%s_%s_hs' % (cold, hot)
                if ID in attempts or (avoid_recycle and match in success): continue
                attempts.add(ID)
                original_cold_stream = get_stream_at_H_max(cold)
                T_cold_in = original_cold_stream.T
                T_hot_in = get_T_transient_hot_side(hot)
                if (Q_hot_side[cold][0]=='heat' and Q_hot_side[cold][1]>0 and
                        T_hot_in - T_cold_in>= T_min_app):    
                    cold_stream = original_cold_stream
                    hot_stream = streams_transient_hot_side[hot].copy()
                    cold_stream.ID = 's_%s__%s'%(cold,ID)
                    hot_stream.ID = 's_%s__%s'%(hot,ID)
                    hot_out = hot_stream.copy('%s__s_%s'%(ID,hot))
                    cold_out = cold_stream.copy('%s__s_%s'%(ID,cold))
                    H_lim = H_out_arr[cold]
                    new_HX = bst.units.HXprocess(ID = ID, ins = (cold_stream, hot_stream),
                             outs = (cold_out, hot_out), H_lim0 = H_lim, 
                             T_lim1 = T_out_arr[hot], dT = T_min_app,
                             thermo = hot_stream.thermo)
                    try: new_HX._run()
                    except: continue
                    if abs(new_HX.Q )< Qmin: continue
                    success.add(match)
                    HXs_hot_side.append(new_HX)                        
                    stream_HXs_dict[hot].append(new_HX)
                    stream_HXs_dict[cold].append(new_HX)                        
                    Q_hot_side[hot][1] -= new_HX.Q
                    Q_hot_side[cold][1] -= new_HX.Q       
                    streams_transient_hot_side[cold] = new_HX.outs[0]
                    streams_transient_hot_side[hot] = new_HX.outs[1]
                    H_out = new_HX.outs[0].H
                    assert H_out - new_HX.ins[0].H >= 0.
                    stream_quenched = H_out > H_lim or np.allclose(H_out, H_lim)                   
                    matches_hs[cold].append(hot)                    
                    if stream_quenched:
                        break
    
    # Add final utility HXs
    new_HX_utils = []    
    for hot in hot_indices:
        hot_stream = get_stream_at_H_min(hot)
        ID = 'Util_%s_cs'%(hot)
        hot_stream.ID = 's_%s__%s'%(hot,ID)
        outlet = hot_stream.copy('%s__s_%s'%(ID,hot))
        new_HX_util = bst.units.HXutility(ID = ID, ins = hot_stream, outs = outlet,
                                          H = H_out_arr[hot], rigorous = True,
                                          thermo = hot_stream.thermo)
        new_HX_util._run()
        s_out = new_HX_util-0
        np.testing.assert_allclose(s_out.H, H_out_arr[hot], rtol=5e-3, atol=1.)
        atol_T = 5. if 's' in hxs[hot].outs[0].phases else 0.001
        np.testing.assert_allclose(s_out.T, T_out_arr[hot], rtol=5e-3, atol=atol_T)
        new_HX_utils.append(new_HX_util)
        stream_HXs_dict[hot].append(new_HX_util)
            
    for cold in cold_indices:
        cold_stream = get_stream_at_H_max(cold)
        ID = 'Util_%s_hs'%(cold)
        cold_stream.ID = 's_%s__%s'%(cold,ID)
        outlet = cold_stream.copy('%s__s_%s'%(ID,cold))
        new_HX_util = bst.units.HXutility(ID = ID, ins = cold_stream, outs = outlet,
                                          H = H_out_arr[cold], rigorous = True,
                                          thermo = cold_stream.thermo)
        new_HX_util._run()
        s_out = new_HX_util.outs[0]
        np.testing.assert_allclose(s_out.H, H_out_arr[cold], rtol=1e-2, atol=1.)
        atol_T = 5. if 's' in hxs[cold].outs[0].phases else 0.001
        np.testing.assert_allclose(s_out.T, T_out_arr[cold], rtol=5e-2, atol=atol_T)
        new_HX_utils.append(new_HX_util)
        stream_HXs_dict[cold].append(new_HX_util)
    
    return HXs_hot_side, HXs_cold_side, new_HX_utils, hxs, T_in_arr,\
           T_out_arr, pinch_T_arr, C_flow_vector, hx_utils_rearranged, streams_inlet, stream_HXs_dict,\
           hot_indices, cold_indices

